"""
The file `integrators.py` contains the definition of all the integrator types,
with a propagate function that is a wrapper to the `propagate` function of each of the
different elements in the beamline. Each integrator has a dictionary with the different elements
it can be used with, so that if this integrator is selected, the propagate function of the element
will be called via the propagate function of the integrator to allow the tracking.

"""
from typing import List, Tuple

import numpy as _np
from numba.typed import List as nList

from .kernels import batched_vector_matrix, batched_vector_matrix_tensor
from .maps import (
    compute_mad_combined_dipole_matrix,
    compute_mad_combined_dipole_tensor,
    compute_mad_drift_matrix,
    compute_mad_drift_tensor,
    compute_mad_quadrupole_matrix,
    compute_mad_quadrupole_tensor,
    compute_transport_combined_dipole_ex_matrix,
    compute_transport_combined_dipole_ex_tensor,
    compute_transport_combined_dipole_matrix,
    compute_transport_combined_dipole_tensor,
    compute_transport_fringe_in_ex_matrix,
    compute_transport_fringe_in_ex_tensor,
    compute_transport_fringe_in_matrix,
    compute_transport_fringe_in_tensor,
    compute_transport_fringe_out_ex_matrix,
    compute_transport_fringe_out_ex_tensor,
    compute_transport_fringe_out_matrix,
    compute_transport_fringe_out_tensor,
    compute_transport_multipole_ex_matrix,
    compute_transport_multipole_ex_tensor,
    compute_transport_multipole_matrix,
    compute_transport_multipole_tensor,
    compute_transport_quadrupole_ex_matrix,
    compute_transport_quadrupole_ex_tensor,
    compute_transport_quadrupole_matrix,
    compute_transport_quadrupole_tensor,
    compute_transport_sextupole_ex_matrix,
    compute_transport_sextupole_ex_tensor,
    compute_transport_sextupole_matrix,
    compute_transport_sextupole_tensor,
    track_madx_bend,
    track_madx_dipedge,
    track_madx_drift,
    track_madx_drift_paraxial,
    track_madx_kicker,
    track_madx_quadrupole,
    track_madx_srotation,
)

__all__ = [
    "IntegratorType",
    "Integrator",
    "MadXIntegrator",
    "MadXParaxialDriftIntegrator",
    "Mad8Integrator",
    "Mad8FirstOrderTaylorIntegrator",
    "Mad8SecondOrderTaylorIntegrator",
    "TransportIntegrator",
    "TransportFirstOrderTaylorIntegrator",
    "TransportFirstOrderTaylorIntegratorExact",
    "TransportSecondOrderTaylorIntegrator",
    "TransportSecondOrderTaylorIntegratorExact",
    "PTCIntegrator",
]


class IntegratorType(type):
    pass


class Integrator(metaclass=IntegratorType):
    @classmethod
    def propagate(
        cls,
        element,
        beam_in: _np.ndarray,
        beam_out: _np.ndarray,
        global_parameters: nList,
    ) -> Tuple[_np.ndarray, _np.ndarray]:
        return beam_in, beam_out


class MadXIntegrator(Integrator):
    METHODS = {
        "DIPEDGE": track_madx_dipedge,
        "RBEND": track_madx_bend,
        "SBEND": track_madx_bend,
        "DRIFT": track_madx_drift,
        "GAP": track_madx_drift,
        "RECTANGULARCOLLIMATOR": track_madx_drift,
        "ELLIPTICALCOLLIMATOR": track_madx_drift,
        "CIRCULARCOLLIMATOR": track_madx_drift,
        "DUMP": track_madx_drift,
        "QUADRUPOLE": track_madx_quadrupole,
        "SROTATION": track_madx_srotation,
        "KICKER": track_madx_kicker,
        "HKICKER": track_madx_kicker,
        "VKICKER": track_madx_kicker,
    }

    @classmethod
    def propagate(cls, element, beam_in: _np.ndarray, beam_out: _np.ndarray, global_parameters: nList):
        return cls.METHODS.get(element.__class__.__name__.upper())(
            beam_in,
            beam_out,
            element.cache,
            global_parameters,
        )

    @classmethod
    def cache(cls, element) -> List:
        return element.parameters


class MadXParaxialDriftIntegrator(MadXIntegrator):
    METHODS = {
        "DIPEDGE": track_madx_dipedge,
        "RBEND": track_madx_bend,
        "SBEND": track_madx_bend,
        "DRIFT": track_madx_drift_paraxial,
        "RECTANGULARCOLLIMATOR": track_madx_drift_paraxial,
        "ELLIPTICALCOLLIMATOR": track_madx_drift_paraxial,
        "CIRCULARCOLLIMATOR": track_madx_drift_paraxial,
        "DUMP": track_madx_drift_paraxial,
        "QUADRUPOLE": track_madx_quadrupole,
        "SROTATION": track_madx_srotation,
        "KICKER": track_madx_kicker,
        "HKICKER": track_madx_kicker,
        "VKICKER": track_madx_kicker,
    }


class Mad8Integrator(Integrator):
    pass


class Mad8FirstOrderTaylorIntegrator(Mad8Integrator):
    MATRICES = {
        "DRIFT": compute_mad_drift_matrix,
        "BEND": compute_mad_combined_dipole_matrix,
        "SBEND": compute_mad_combined_dipole_matrix,
        "QUADRUPOLE": compute_mad_quadrupole_matrix,
    }

    @classmethod
    def propagate(cls, element, beam_in, beam_out, global_parameters: nList):
        if element.__class__.__name__.upper() in ["HKICKER", "VKICKER"]:
            return track_madx_kicker(beam_in, beam_out, element.cache, global_parameters)

        elif element.__class__.__name__.upper() in [
            "GAP",
            "RECTANGULARCOLLIMATOR",
            "ELLIPTICALCOLLIMATOR",
            "CIRCULARCOLLIMATOR",
            "DUMP",
        ]:
            return batched_vector_matrix(
                beam_in,
                beam_out,
                compute_mad_drift_matrix(element.cache, global_parameters),
            )

        elif element.__class__.__name__.upper() == "SROTATION":
            return track_madx_srotation(beam_in, beam_out, element.cache, global_parameters)

        else:
            return batched_vector_matrix(
                beam_in,
                beam_out,
                cls.MATRICES.get(element.__class__.__name__.upper())(element.cache, global_parameters),
            )

    @classmethod
    def cache(cls, element) -> List:
        return element.parameters


class Mad8SecondOrderTaylorIntegrator(Mad8FirstOrderTaylorIntegrator):
    TENSORS = {
        "DRIFT": compute_mad_drift_tensor,
        "BEND": compute_mad_combined_dipole_tensor,
        "SBEND": compute_mad_combined_dipole_tensor,
        "QUADRUPOLE": compute_mad_quadrupole_tensor,
    }

    @classmethod
    def propagate(cls, element, beam_in, beam_out, global_parameters: nList):
        if element.__class__.__name__.upper() in ["HKICKER", "VKICKER"]:
            return track_madx_kicker(beam_in, beam_out, element.cache, global_parameters)

        elif element.__class__.__name__.upper() in [
            "GAP",
            "RECTANGULARCOLLIMATOR",
            "ELLIPTICALCOLLIMATOR",
            "CIRCULARCOLLIMATOR",
            "DUMP",
        ]:
            return batched_vector_matrix_tensor(
                beam_in,
                beam_out,
                compute_mad_drift_matrix(element.cache, global_parameters),
                compute_mad_drift_tensor(element.cache, global_parameters),
            )

        elif element.__class__.__name__.upper() == "SROTATION":
            return track_madx_srotation(beam_in, beam_out, element.cache, global_parameters)

        else:
            return batched_vector_matrix_tensor(
                beam_in,
                beam_out,
                cls.MATRICES.get(element.__class__.__name__.upper())(element.cache, global_parameters),
                cls.TENSORS.get(element.__class__.__name__.upper())(element.cache, global_parameters),
            )

    @classmethod
    def cache(cls, element) -> List:
        return element.parameters


class TransportIntegrator(Integrator):
    pass


class TransportFirstOrderTaylorIntegrator(TransportIntegrator):
    MATRICES = {
        "BEND": compute_transport_combined_dipole_matrix,
        "SBEND": compute_transport_combined_dipole_matrix,
        "QUADRUPOLE": compute_transport_quadrupole_matrix,
        "SEXTUPOLE": compute_transport_sextupole_matrix,
        "MULTIPOLE": compute_transport_multipole_matrix,
        "FRINGEIN": compute_transport_fringe_in_matrix,
        "FRINGEOUT": compute_transport_fringe_out_matrix,
    }

    @classmethod
    def propagate(cls, element, beam_in, beam_out, global_parameters: nList):
        if element.__class__.__name__.upper() in ["HKICKER", "VKICKER"]:
            return track_madx_kicker(beam_in, beam_out, element.cache, global_parameters)

        elif element.__class__.__name__.upper() in [
            "DRIFT",
            "GAP",
            "RECTANGULARCOLLIMATOR",
            "ELLIPTICALCOLLIMATOR",
            "CIRCULARCOLLIMATOR",
            "DUMP",
        ]:
            return track_madx_drift(beam_in, beam_out, element.cache, global_parameters)

        elif element.__class__.__name__.upper() == "SROTATION":
            return track_madx_srotation(beam_in, beam_out, element.cache, global_parameters)

        else:
            b = batched_vector_matrix(
                beam_in,
                beam_out,
                cls.MATRICES.get(element.__class__.__name__.upper())(element.cache),
            )
            return b[0], b[1]

    @classmethod
    def cache(cls, element) -> List:
        return element.parameters


class TransportFirstOrderTaylorIntegratorExact(TransportIntegrator):
    MATRICES = {
        "BEND": compute_transport_combined_dipole_ex_matrix,
        "SBEND": compute_transport_combined_dipole_ex_matrix,
        "QUADRUPOLE": compute_transport_quadrupole_ex_matrix,
        "SEXTUPOLE": compute_transport_sextupole_ex_matrix,
        "MULTIPOLE": compute_transport_multipole_ex_matrix,
        "FRINGEIN": compute_transport_fringe_in_ex_matrix,
        "FRINGEOUT": compute_transport_fringe_out_ex_matrix,
    }

    @classmethod
    def propagate(cls, element, beam_in, beam_out, global_parameters: nList):
        if element.__class__.__name__.upper() in ["HKICKER", "VKICKER"]:
            return track_madx_kicker(beam_in, beam_out, element.cache, global_parameters)

        elif element.__class__.__name__.upper() in [
            "DRIFT",
            "GAP",
            "RECTANGULARCOLLIMATOR",
            "ELLIPTICALCOLLIMATOR",
            "CIRCULARCOLLIMATOR",
            "DUMP",
        ]:
            return track_madx_drift(beam_in, beam_out, element.cache, global_parameters)

        elif element.__class__.__name__.upper() == "SROTATION":
            return track_madx_srotation(beam_in, beam_out, element.cache, global_parameters)

        else:
            b2 = _np.zeros(beam_in.shape)
            for i in range(beam_in.shape[0]):
                updated_parameters = element.cache.copy()

                updated_parameters.append(beam_in[i, 4])
                matrix = cls.MATRICES.get(element.__class__.__name__.upper())(updated_parameters)
                b2[i, :] = batched_vector_matrix(_np.array([beam_in[i, :]]), _np.array([beam_out[i, :]]), matrix)[1]

            return beam_in, b2

    @classmethod
    def cache(cls, element) -> List:
        return element.parameters


class TransportSecondOrderTaylorIntegrator(TransportFirstOrderTaylorIntegrator):
    TENSORS = {
        "BEND": compute_transport_combined_dipole_tensor,
        "SBEND": compute_transport_combined_dipole_tensor,
        "QUADRUPOLE": compute_transport_quadrupole_tensor,
        "SEXTUPOLE": compute_transport_sextupole_tensor,
        "MULTIPOLE": compute_transport_multipole_tensor,
        "FRINGEIN": compute_transport_fringe_in_tensor,
        "FRINGEOUT": compute_transport_fringe_out_tensor,
    }

    @classmethod
    def propagate(cls, element, beam_in, beam_out, global_parameters: nList):
        if element.__class__.__name__.upper() in ["HKICKER", "VKICKER"]:
            return track_madx_kicker(beam_in, beam_out, element.cache, global_parameters)

        elif element.__class__.__name__.upper() in [
            "DRIFT",
            "GAP",
            "RECTANGULARCOLLIMATOR",
            "ELLIPTICALCOLLIMATOR",
            "CIRCULARCOLLIMATOR",
            "DUMP",
        ]:
            return track_madx_drift(beam_in, beam_out, element.cache, global_parameters)

        elif element.__class__.__name__.upper() == "SROTATION":
            return track_madx_srotation(beam_in, beam_out, element.cache, global_parameters)

        else:
            b = batched_vector_matrix_tensor(
                beam_in,
                beam_out,
                cls.MATRICES.get(element.__class__.__name__.upper())(element.cache),
                cls.TENSORS.get(element.__class__.__name__.upper())(element.cache),
            )
            return b[0], b[1]

    @classmethod
    def cache(cls, element) -> List:
        return element.parameters


class TransportSecondOrderTaylorIntegratorExact(TransportFirstOrderTaylorIntegratorExact):
    TENSORS = {
        "BEND": compute_transport_combined_dipole_ex_tensor,
        "SBEND": compute_transport_combined_dipole_ex_tensor,
        "QUADRUPOLE": compute_transport_quadrupole_ex_tensor,
        "SEXTUPOLE": compute_transport_sextupole_ex_tensor,
        "MULTIPOLE": compute_transport_multipole_ex_tensor,
        "FRINGEIN": compute_transport_fringe_in_ex_tensor,
        "FRINGEOUT": compute_transport_fringe_out_ex_tensor,
    }

    @classmethod
    def propagate(cls, element, beam_in, beam_out, global_parameters: nList):
        if element.__class__.__name__.upper() in ["HKICKER", "VKICKER"]:
            return track_madx_kicker(beam_in, beam_out, element.cache, global_parameters)

        elif element.__class__.__name__.upper() in [
            "DRIFT",
            "GAP",
            "RECTANGULARCOLLIMATOR",
            "ELLIPTICALCOLLIMATOR",
            "CIRCULARCOLLIMATOR",
            "DUMP",
        ]:
            return track_madx_drift(beam_in, beam_out, element.cache, global_parameters)

        elif element.__class__.__name__.upper() == "SROTATION":
            return track_madx_srotation(beam_in, beam_out, element.cache, global_parameters)

        else:
            b2 = _np.zeros(beam_in.shape)
            for i in range(beam_in.shape[0]):
                updated_parameters = element.cache.copy()
                updated_parameters.append(beam_in[i, 5])
                matrix = cls.MATRICES.get(element.__class__.__name__.upper())(updated_parameters)
                tensor = cls.TENSORS.get(element.__class__.__name__.upper())(updated_parameters)
                b2[i, :] = batched_vector_matrix_tensor(
                    _np.array([beam_in[i, :]]),
                    _np.array([beam_out[i, :]]),
                    matrix,
                    tensor,
                )[1]

            return beam_in, b2

    @classmethod
    def cache(cls, element) -> List:
        return element.parameters


class PTCIntegrator(Integrator):
    pass



import numpy as np
from numba import njit
from numba.typed import List as nList
from numpy import cos, sin, tan


@njit(cache=True)
def compute_transport_fringe_out_ex_matrix(element_parameters: nList) -> np.ndarray:
    L: float = element_parameters[0]
    alpha: float = element_parameters[1]
    beta2: float = element_parameters[3]
    gap: float = element_parameters[4]
    fintx: float = element_parameters[5]
    d: float = element_parameters[len(element_parameters) - 1]

    h = alpha / L
    R = np.zeros((6, 6))
    h = h / (1 + d)
    phi2 = fintx * h * gap * cos(beta2) ** (-1) * (1 + sin(beta2) ** 2)
    R[0, 0] = 1
    R[1, 0] = h * tan(beta2)
    R[1, 1] = 1
    R[2, 2] = 1
    R[3, 2] = -h * tan(beta2 - phi2)
    R[3, 3] = 1
    R[4, 4] = 1
    R[5, 5] = 1
    return R


@njit(cache=True)
def compute_transport_fringe_out_ex_tensor(element_parameters: nList) -> np.ndarray:
    L: float = element_parameters[0]
    alpha: float = element_parameters[1]
    k1: float = element_parameters[2]
    beta2: float = element_parameters[3]
    gap: float = element_parameters[4]
    fintx: float = element_parameters[5]
    r2: float = element_parameters[6]
    d: float = element_parameters[len(element_parameters) - 1]
    h = alpha / L

    T = np.zeros((6, 6, 6))
    h = h / (1 + d)
    k1 = k1 / (1 + d)
    phi2 = fintx * h * gap * cos(beta2) ** (-1) * (1 + sin(beta2) ** 2)
    T[0, 0, 0] = h * tan(beta2) ** 2 / 2
    T[0, 2, 2] = -h * cos(beta2) ** (-1) ** 2 / 2
    T[1, 0, 0] = k1 * tan(beta2) - 0.5 * h**2 * tan(beta2) ** 3 + h * cos(beta2) ** (-1) ** 3 / (2 * r2)
    T[1, 0, 1] = -h * tan(beta2) ** 2
    T[1, 0, 4] = -h * tan(beta2)
    T[1, 2, 2] = -k1 * tan(beta2) - 0.5 * h**2 * tan(beta2) ** 3 - h * cos(beta2) ** (-1) ** 3 / (2 * r2)
    T[1, 2, 3] = h * tan(beta2) ** 2
    T[2, 0, 2] = -h * tan(beta2) ** 2
    T[3, 0, 2] = -2 * k1 * tan(beta2) + h**2 * tan(beta2) * cos(beta2) ** (-1) ** 2 - h * cos(beta2) ** (-1) ** 3 / r2
    T[3, 0, 3] = h * tan(beta2) ** 2
    T[3, 1, 2] = h * cos(beta2) ** (-1) ** 2
    T[3, 2, 4] = -h * phi2 * cos(beta2 - phi2) ** (-2) + h * tan(beta2)
    return T



from .beam import Beam
from .core import match, track, twiss
from .elements import (
    BeamStop,
    Bend,
    CircularCollimator,
    Degrader,
    DipEdge,
    Drift,
    Fringein,
    Fringeout,
    HKicker,
    Kicker,
    Marker,
    Multipole,
    Quadrupole,
    RBend,
    RectangularCollimator,
    SBend,
    Scatterer,
    Sextupole,
    SRotation,
    TKicker,
    VKicker,
)
from .input import Input
from .integrators import *
from .observers import BeamObserver, IbaBpmObserver, LossesObserver, MeanObserver, Observer, SigmaObserver



"""
The file `input.py` contains the main class that needs instantiated to build a Manzoni input.
It requires at least a beamline Sequence and a Beam instance to allow the tracking. Once instantiated,
the Input class has a track function, which is just a wrapper on the track function of the file `core.py`,
and can be directly called to perform the tracking through the beamline. There are also a couple of
useful functions to freeze, unfreeze, set or get the parameters of a given element of the sequence.
Finally, the “adjust_energy” function is important to calculate all the initial kinetic energies of
all the scatterers and degraders of the beamline, starting from the initial kinetic energy of the beam.
It is necessary to correct the particles' propagation through these elements during the tracking.
"""
from __future__ import annotations

from typing import Dict, List, Optional, Union

import numpy as _np
import pandas as _pd
from georges_core import Kinematics as _Kinematics
from georges_core import ureg as _ureg
from georges_core.sequences import BetaBlock as _BetaBlock
from georges_core.sequences import Sequence as _Sequence

from ..fermi import materials
from . import elements
from .beam import Beam as _Beam
from .core import track, twiss
from .elements import ManzoniElement
from .elements.scatterers import MaterialElement
from .integrators import Integrator, MadXIntegrator
from .observers import Observer as _Observer

MANZONI_FLAVOR = {"Sbend": "SBend", "Rbend": "RBend"}


class Input:
    def __init__(
        self,
        sequence: Optional[List[elements.ManzoniElement]] = None,
        beam: Optional[_Beam] = None,
        mapper: Dict[str, int] = None,
    ):
        self._sequence = sequence
        self._beam = beam
        self.mapper = mapper

    @property
    def sequence(self):  # pragma: no cover
        return self._sequence

    @property
    def beam(self):  # pragma: no cover
        return self._beam

    def to_df(self):
        """

        Returns: A pandas.DataFrame of the sequence

        """

        _ = list(map(lambda e: _pd.Series(e.attributes), self.sequence))
        df = _pd.concat(_, axis=1).T
        df["CLASS"] = list(map(lambda e: e.__class__.__name__, self.sequence))
        return df.set_index("NAME")

    @property
    def df(self):  # pragma: no cover
        return self.to_df()

    def freeze(self):
        """
        Freezes all elements in the input sequence.

        Returns:
            `self` to allow method chaining
        """
        for e in self._sequence:
            e.freeze()
        return self

    def unfreeze(self):
        """
        Unfreezes all elements in the input sequence.

        Returns:
            `self` to allow method chaining
        """
        for e in self._sequence:
            e.unfreeze()
        return self

    def track(
        self,
        beam: _Beam,
        observers: Union[List[_Observer], _Observer] = None,
        check_apertures: bool = True,
    ) -> Union[List[_Observer], _Observer]:
        """

        Args:
            beam:
            observers:
            check_apertures:

        Returns:
            the `Observer` object containing the tracking results.
        """
        if not isinstance(observers, list):
            observers = [observers]
        track(self, beam, observers, check_apertures_exit=check_apertures)
        if observers is not None:
            if len(observers) == 1:
                return observers[0]
            else:
                return observers

    def twiss(
        self,
        kinematics: _Kinematics,
        reference_particle: _np.ndarray = None,
        offsets=None,
        twiss_parametrization: bool = True,
        twiss_init: _BetaBlock = None,
    ) -> _pd.DataFrame:
        """

        Args:
            kinematics:
            reference_particle:
            offsets:
            twiss_parametrization:
            twiss_init:

        Returns:
            The dataframe with the Twiss functions at each element.

        """
        return twiss(self, kinematics, reference_particle, offsets, twiss_parametrization, twiss_init)

    def adjust_energy(self, input_energy: _ureg.Quantity):
        current_energy = input_energy
        for e in self.sequence:
            if isinstance(e, MaterialElement):
                e.KINETIC_ENERGY = current_energy
                current_energy = e.degraded_energy

    def compute_efficiency(self, input_energy: _ureg.Quantity) -> float:
        self.adjust_energy(input_energy)
        efficiency = 1.0
        for e in self._sequence:
            if isinstance(e, MaterialElement):
                efficiency *= e.cache[5]
        return efficiency

    def set_integrator(self, integrator: Integrator = MadXIntegrator):
        for i in range(len(self.mapper)):
            self.sequence[i].integrator = integrator

    # TODO: use method __setitem__ instead ?
    def set_parameters(self, element: str, parameters: Dict):
        # unfreeze the element
        self.sequence[self.mapper[element]].unfreeze()
        for param in parameters.keys():
            self.sequence[self.mapper[element]].__setattr__(param, parameters[param])
        self.sequence[self.mapper[element]].freeze()

    def get_parameters(self, element: str, parameters: Optional[List] = None):
        if parameters is None:
            parameters = self.sequence[self.mapper[element]].attributes
        return dict(zip(parameters, list(map(self.sequence[self.mapper[element]].__getattr__, parameters))))

    @classmethod
    def insert_thin_element(
        cls,
        sequence: Input = None,
        position: int = 0,
        thin_element: ManzoniElement = None,
    ) -> Input:
        """
        Insert a thin element (e.g fringes) in the sequence.

        Args:
            sequence: the manzoni sequence
            position: index of the position to insert in the sequence
            thin_element: element to insert

        Returns:
            A new instance of manzoni.Input
        """

        new_sequence = sequence.sequence.copy()
        new_sequence.insert(position, thin_element)
        element_mapper = {k.NAME: v for v, k in enumerate(new_sequence)}
        return cls(sequence=new_sequence, mapper=element_mapper)

    @classmethod
    def from_sequence(
        cls,
        sequence: _Sequence,
        from_element: str = None,
        to_element: str = None,
    ):
        """
        Creates a new `Input` from a generic sequence from `georges_core`.

        Args:
            sequence:
            from_element:
            to_element:
        Returns:

        """

        input_sequence = list()
        df_sequence = sequence.df.loc[from_element:to_element]
        if "MATERIAL" in df_sequence.columns:
            idx = df_sequence[sequence.df["MATERIAL"].notnull()].index
            for ele in idx:
                if not isinstance(df_sequence.loc[ele, "MATERIAL"], materials.CompoundType):
                    df_sequence.loc[ele, "MATERIAL"] = getattr(materials, df_sequence.loc[ele, "MATERIAL"])

        for name, element in df_sequence.iterrows():
            element = element.replace({_np.nan: None})
            element_class = getattr(elements, MANZONI_FLAVOR.get(element["CLASS"], element["CLASS"]))
            parameters = list(set(list(element.index.values)).intersection(element_class.PARAMETERS.keys()))
            input_sequence.append(
                element_class(name, **element[parameters]),
            )
        element_mapper = {k: v for v, k in enumerate(list(df_sequence.index.values))}
        return cls(sequence=input_sequence, mapper=element_mapper)



"""Matplotlib plotting module for Manzoni.

TODO
"""

from __future__ import annotations

import logging
from typing import Union

import cpymad.madx
import matplotlib.patches as patches
import matplotlib.ticker as mticker
import numpy as _np
import pandas as _pd
from georges_core.vis import MatplotlibArtist as _MatplotlibArtist
from georges_core.vis.artist import PALETTE
from lmfit.models import GaussianModel

from ..manzoni.observers import BeamObserver as _BeamObserver
from ..manzoni.observers import IbaBpmObserver as _IbaBpmObserver
from ..manzoni.observers import LossesObserver as _LossesObserver
from ..manzoni.observers import MeanObserver as _MeanObserver
from ..manzoni.observers import Observer as _Observer
from ..manzoni.observers import SigmaObserver as _SigmaObserver
from ..manzoni.observers import SymmetryObserver as _SymmetryObserver
from ..manzoni.observers import TwissObserver as _TwissObserver

palette = PALETTE["solarized"]
palette["both"] = palette["base03"]
palette["X"] = palette["cyan"]
palette["Y"] = palette["orange"]
palette["XP"] = palette["red"]
palette["YP"] = palette["green"]
palette["TWISS_X"] = palette["blue"]
palette["TWISS_Y"] = palette["red"]
palette["DX"] = palette["green"]
palette["DY"] = palette["orange"]


# THIS IS THE OLD beam.py
class BeamPlottingException(Exception):
    """Exception raised for errors in the beam plotting module."""

    def __init__(self, m):
        self.message = m


class StatisticException(Exception):
    """Exception raised for errors in the beam plotting module."""

    def __init__(self, m):
        self.message = m


class ManzoniMatplotlibArtist(_MatplotlibArtist):
    """A matplotlib implementation of a `Matplotlib` artist."""

    def __init__(self, tracks_color: str = "b", **kwargs):
        """
        Args:
            param ax: the matplotlib ax used for plotting. If None it will be created with `init_plot` (kwargs are
            forwarded).
            tracks_color: color for the plotting of tracks
            kwargs: forwarded to `MatplotlibPlot` and to `init_plot`.
        """
        super().__init__(**kwargs)
        self._tracks_color = tracks_color

    @property
    def tracks_color(self):
        """
        The color for the rendering of the tracks.

        Returns:
            color as a string
        """
        return self._tracks_color

    @tracks_color.setter
    def tracks_color(self, color: str):
        self._tracks_color = color

    @staticmethod
    def histogram_fit(data, bounds_binning=50, verbose=False, model=GaussianModel):
        """All models are available on https://lmfit.github.io/lmfit-py/builtin_models.html#lmfit.models"""
        y, bin_edges = _np.histogram(data, density=False, bins=bounds_binning)
        x = (bin_edges[:-1] + bin_edges[1:]) / 2
        result = model().fit(data=y, x=x, center=data.mean(), sigma=data.std())
        if verbose:
            print(result.fit_report())
        return x, result

    @staticmethod
    def ellipse(ra, rb, angle, x0, y0, **kwargs):
        """
        Create an ellipse from beam parameters.
        :param ra: semi-major axis
        :param rb: semi-minor axis
        :param angle: oritentation angle
        :param x0: center X coordinate
        :param y0: center Y coordinate
        :return: `matplotlib.patches.Ellipse` object
        """
        width = ra
        height = rb
        return patches.Ellipse(
            (x0, y0),
            width,
            height,
            angle=angle,
            linewidth=kwargs.get("linewidth", 2),
            fill=kwargs.get("fill", False),
            linestyle=kwargs.get("linestyle", "--"),
            edgecolor=kwargs.get("color", "red"),
            label=kwargs.get("label"),
        )

    @staticmethod
    def rotation_angle(e_val, evec):
        if e_val[0] * evec[0, 0] > e_val[1] * evec[0, 1]:
            return _np.degrees(_np.arctan(evec[1, 0] / evec[0, 0]))
        else:
            return _np.degrees(_np.arctan(evec[1, 1] / evec[0, 1]))

    # Plotting for the tracking
    def tracking(
        self,
        observer: _Observer = None,
        plane: str = "X",
        fill_between: bool = False,
        mean: bool = True,
        std: bool = False,
        halo: bool = True,
        **kwargs,
    ):
        """
        Plot the beam envelopes from tracking data.

        Args:
            observer: Observer used for the tracking
            plane:
            fill_between:
            mean:
            std:
            halo:
            **kwargs:

        Returns:

        """

        tracking_palette = kwargs.get("palette", palette)
        df_observer = observer.to_df()

        if isinstance(observer, _MeanObserver):
            self._ax.plot(
                _np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values]),
                _np.hstack(
                    [
                        df_observer.iloc[0][f"BEAM_IN_{plane}"],
                        df_observer.iloc[:][f"BEAM_OUT_{plane}"].values,
                    ],
                )
                * 1000,
                "^-",
                color=tracking_palette[plane],
                markeredgecolor=tracking_palette[plane],
                markersize=2,
                linewidth=1,
                label=kwargs.get("label", plane),
            )
            self._ax.set_xlabel("S (m)")
            self._ax.set_ylabel("Mean Position (mm)")

        elif isinstance(observer, _SigmaObserver):
            x = _np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values])
            if plane == "both":
                y0 = (
                    _np.hstack(
                        [
                            df_observer.iloc[0]["BEAM_IN_X"],
                            df_observer.iloc[:]["BEAM_OUT_X"].values,
                        ],
                    )
                    * 1000
                )
                y1 = (
                    _np.hstack(
                        [
                            df_observer.iloc[0]["BEAM_IN_Y"],
                            df_observer.iloc[:]["BEAM_OUT_Y"].values,
                        ],
                    )
                    * 1000
                )
            else:
                y0 = (
                    _np.hstack(
                        [
                            df_observer.iloc[0][f"BEAM_IN_{plane}"],
                            df_observer.iloc[:][f"BEAM_OUT_{plane}"].values,
                        ],
                    )
                    * 1000
                )
                y1 = y0
            y = [y1, -y0]

            if plane == "both":
                label = ""
            else:
                label = kwargs.get("label", plane)
            self._ax.plot(
                x,
                y[0],
                x,
                y[1],
                marker="^",
                color=tracking_palette[plane],
                markeredgecolor=tracking_palette[plane],
                markersize=2,
                linewidth=1,
                label=label,
            )
            if fill_between:
                self._ax.fill_between(
                    x,
                    y[1],
                    y[0],
                    facecolor=tracking_palette[plane],
                    linewidth=0.0,
                    edgecolor=tracking_palette[plane],
                )
            self._ax.set_xlabel("S (m)")
            self._ax.set_ylabel("Beam Size (mm)")

            if plane == "both":
                self._ax.get_lines()[0].set_color(tracking_palette["Y"])
                self._ax.get_lines()[1].set_color(tracking_palette["X"])
                ticks_loc = self._ax.get_yticks().tolist()
                self._ax.yaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
                self._ax.set_yticklabels([abs(x) for x in ticks_loc])

                self._ax.annotate(
                    "",
                    xy=(-0.08 / (self._ax.get_figure().get_size_inches()[0] / 10), 0.97),
                    xytext=(-0.08 / (self._ax.get_figure().get_size_inches()[0] / 10), 0.75),
                    arrowprops=dict(arrowstyle="->", color="k"),
                    xycoords=self._ax.transAxes,
                )

                self._ax.annotate(
                    "",
                    xy=(-0.08 / (self._ax.get_figure().get_size_inches()[0] / 10), 0.25),
                    xycoords="axes fraction",
                    xytext=(-0.08 / (self._ax.get_figure().get_size_inches()[0] / 10), 0.03),
                    arrowprops=dict(arrowstyle="<-", color="k"),
                )

                self._ax.text(
                    -0.1 / (self._ax.get_figure().get_size_inches()[0] / 10),
                    0.86,
                    "Vertical",
                    fontsize=9,
                    rotation=90,
                    transform=self._ax.transAxes,
                )
                self._ax.text(
                    -0.1 / (self._ax.get_figure().get_size_inches()[0] / 10),
                    0.11,
                    "Horizontal",
                    fontsize=9,
                    rotation=90,
                    transform=self._ax.transAxes,
                )

        elif isinstance(observer, _BeamObserver):
            dico_plane = {"X": 0, "PX": 1, "Y": 2, "PY": 3}
            t = df_observer.apply(
                lambda r: _pd.Series(
                    {
                        "S": r["AT_EXIT"].m_as("m"),
                        "mean": 1000 * r["BEAM_OUT"][:, dico_plane[plane]].mean() if mean else 0.0,
                        "std": 1000 * r["BEAM_OUT"][:, dico_plane[plane]].std() if std else 0.0,
                        "1%": 1000
                        * (
                            self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.023)
                            - self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.5)
                        ),
                        "5%": 1000
                        * (
                            self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.159)
                            - self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.5)
                        ),
                        "20%": 1000
                        * (
                            self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 1 - 0.842701)
                            - self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.5)
                        ),
                        "80%": 1000
                        * (
                            self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.842701)
                            - self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.5)
                        ),
                        "95%": 1000
                        * (
                            self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.841)
                            - self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.5)
                        ),
                        "99%": 1000
                        * (
                            self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.977)
                            - self.compute_halo(r["BEAM_OUT"][:, dico_plane[plane]], 0.5)
                        ),
                    },
                ),
                axis=1,
            )

            if df_observer.iloc[0]["BEAM_IN"] is not None:
                data_entry = df_observer.iloc[0]
                t0 = _pd.DataFrame(
                    data={
                        "S": [data_entry["AT_ENTRY"].m_as("m")],
                        "mean": [
                            1000
                            * data_entry["BEAM_IN"][
                                :,
                                dico_plane[plane],
                            ].mean()
                            if mean
                            else 0.0,
                        ],
                        "std": [
                            1000 * data_entry["BEAM_IN"][:, dico_plane[plane]].std() if std else 0.0,
                        ],
                        "1%": [
                            1000
                            * (
                                self.compute_halo(
                                    data_entry["BEAM_IN"][:, dico_plane[plane]],
                                    0.023,
                                )
                                - self.compute_halo(
                                    data_entry["BEAM_IN"][:, dico_plane[plane]],
                                    0.5,
                                )
                            ),
                        ],
                        "5%": [
                            1000
                            * (
                                self.compute_halo(
                                    data_entry["BEAM_IN"][:, dico_plane[plane]],
                                    0.159,
                                )
                                - self.compute_halo(
                                    data_entry["BEAM_IN"][:, dico_plane[plane]],
                                    0.5,
                                )
                            ),
                        ],
                        "95%": [
                            1000
                            * (
                                self.compute_halo(
                                    data_entry["BEAM_IN"][:, dico_plane[plane]],
                                    0.841,
                                )
                                - self.compute_halo(
                                    data_entry["BEAM_IN"][:, dico_plane[plane]],
                                    0.5,
                                )
                            ),
                        ],
                        "99%": [
                            1000
                            * (
                                self.compute_halo(
                                    data_entry["BEAM_IN"][:, dico_plane[plane]],
                                    0.977,
                                )
                                - self.compute_halo(
                                    data_entry["BEAM_IN"][:, dico_plane[plane]],
                                    0.5,
                                )
                            ),
                        ],
                    },
                    index=["Start"],
                )
                t = _pd.concat([t0, t])

            if t["S"].count == 0:
                return

            if halo:
                self.filled_plot(
                    self._ax,
                    t["S"],
                    t["mean"] + t["5%"],
                    t["mean"] + t["95%"],
                    tracking_palette[plane],
                    fill=True,
                    alpha=0.3,
                )
                self.filled_plot(
                    self._ax,
                    t["S"],
                    t["mean"] + t["1%"],
                    t["mean"] + t["99%"],
                    tracking_palette[plane],
                    fill=True,
                    alpha=0.3,
                )

            if mean:
                self._ax.plot(
                    t["S"],
                    t["mean"],
                    "*-",
                    color=tracking_palette[plane],
                    markeredgecolor=tracking_palette[plane],
                    markersize=2,
                    linewidth=1,
                    label="mean",
                )

            if std:
                self._ax.plot(
                    t["S"],
                    t["mean"] + t["std"],
                    "^-",
                    color=tracking_palette[plane],
                    markeredgecolor=tracking_palette[plane],
                    markersize=2,
                    linewidth=1,
                    label="std",
                )
                self._ax.plot(
                    t["S"],
                    t["mean"] - t["std"],
                    "v-",
                    color=tracking_palette[plane],
                    markeredgecolor=tracking_palette[plane],
                    markersize=2,
                    linewidth=1,
                )

            self._ax.set_xlabel("S (m)")
            self._ax.set_ylabel("Beam Size (mm)")

        elif isinstance(observer, _IbaBpmObserver):
            logging.warning("The plotting method for IbaBpmObserver is not yet implemented.")
            # Adjustment to avoid plotting zero values where no BPM is present
            # if std_bpm:
            #
            #     t.loc[t.std_bpm == 0, 'std_bpm'] = -1000
            #     ax.errorbar(t['S'] - 0.05, t['std_bpm'], xerr=0.1, yerr=t['std_bpm_err'],
            #                 fmt='none',
            #                 elinewidth=2.0,
            #                 linewidth=0.0,
            #                 color=tracking_palette['green'])
            #     ax.errorbar(t['S'] - 0.05, -t['std_bpm'], xerr=0.1, yerr=t['std_bpm_err'],
            #                 fmt='none',
            #                 elinewidth=2.0,
            #                 linewidth=0.0,
            #                 color=tracking_palette['green'])

        elif isinstance(observer, _SymmetryObserver):
            raise BeamPlottingException("Use method vis.ManzoniMatplotlibArtist(ax=ax).symmetry to plot symmetry.")

        elif isinstance(observer, _LossesObserver):
            raise BeamPlottingException("Use method vis.ManzoniMatplotlibArtist(ax=ax).losses to plot losses.")

        else:
            raise BeamPlottingException(f"No plotting method for {observer} is implemented")

    # Plotting for the losses
    def losses(self, observer: _LossesObserver = None, log_scale: bool = False, **kwargs):
        """
        Plot the losses along the beamline

        Args:
            observer: Observer used for the tracking
            log_scale: Log scale for transmission
        """

        if not isinstance(observer, _LossesObserver):
            raise BeamPlottingException("The observer must be a LossesObserver.")

        losses_palette = kwargs.get("palette", palette)
        df_observer = observer.to_df()
        data_center = df_observer["AT_CENTER"].apply(lambda e: e.m_as("m"))

        self._ax.set_xlabel(r"S (m)")
        self._ax.set_ylabel(r"Losses ($\%$)")
        self._ax.yaxis.label.set_color(losses_palette["magenta"])
        self._ax.bar(
            data_center,
            df_observer["LOSSES"],
            width=0.125,
            alpha=0.7,
            edgecolor=losses_palette["magenta"],
            color=losses_palette["magenta"],
            align="center",
            error_kw=dict(ecolor=losses_palette["base02"], capsize=2, capthick=1),
        )
        max_val = (df_observer["LOSSES"]).abs().max()
        self._ax.yaxis.set_major_locator(mticker.MultipleLocator(_np.ceil(max_val / 10)))
        self._ax.set_ylim([0, max_val + 5.0])

        ax2 = self._ax.twinx()  # self.ax_2 is reserved for cartouche
        ax2.set_ylabel(r"T ($\%$)")
        ax2.yaxis.label.set_color(losses_palette["green"])
        ax2.grid(True)
        # TODO USE Transmission, shift and compute ?
        init = df_observer.iloc[0]["PARTICLES_IN"]
        global_transmission = 100 * (df_observer["PARTICLES_OUT"].values / init)
        if log_scale:
            ax2.semilogy(
                _np.hstack([0, data_center.values]),
                _np.hstack([100, global_transmission]),
                "s-",
                color=losses_palette["green"],
            )
            ax2.set_ylim([min(global_transmission), 100])
        else:
            ax2.yaxis.set_major_locator(mticker.MultipleLocator(10))
            ax2.set_ylim([0, 100])
            ax2.plot(
                _np.hstack([0, data_center.values]),
                _np.hstack([100, global_transmission]),
                "s-",
                color=losses_palette["green"],
            )

    def symmetry(self, observer: _LossesObserver = None, **kwargs):
        """
        Plot the symmetry of the beam along the beamline

        Args:
            observer: Observer used for the tracking
        """

        if not isinstance(observer, _SymmetryObserver):
            raise BeamPlottingException("The observer must be a SymmetryObserver.")
        symmetry_palette = kwargs.get("palette", palette)

        df_observer = observer.to_df()
        self._ax.plot(
            _np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values]),
            _np.hstack(
                [
                    df_observer.iloc[0]["SYM_IN"],
                    df_observer.iloc[:]["SYM_OUT"].values,
                ],
            )
            * 100,
            "^-",
            color=symmetry_palette["blue"],
            markeredgecolor=symmetry_palette["blue"],
            markersize=2,
            linewidth=1,
        )
        self._ax.set_xlabel("S (m)")
        self._ax.set_ylabel("Asymmetry ($\%$)")  # noqa: W605

    @staticmethod
    def compute_halo(data, percentile):
        """Return a dataframe containing the 1st, 5th, 95th and 99th percentiles of each dimensions."""
        return _np.quantile(data, percentile)

    @staticmethod
    def filled_plot(ax, x, y0, y, c, fill=False, **kwargs):
        ax.plot(x, y, ".", markersize=0, markerfacecolor=c, markeredgecolor=c, color=c, **kwargs)
        if fill:
            ax.fill_between(x, y0, y, facecolor=c, linewidth=0.0, edgecolor=c, **kwargs)

    # Plotting for the Twiss
    def twiss(
        self,
        observer: _TwissObserver = None,
        with_beta: bool = True,
        with_alpha: bool = False,
        with_dispersion: bool = False,
        tfs_data: Union[_pd.DataFrame, cpymad.madx.Table] = None,
        relativistic_beta: float = 1.0,
        **kwargs,
    ):
        """
        Plot the Twiss function along the beamline

        Args:
            observer: Observer used for the tracking
            with_beta: plot the beta
            with_alpha: plot the alpha
            with_dispersion: plot the dispersion
            tfs_data: if provided, plot the data from MAD-X.
            relativistic_beta (float): Relativistic beta value to scale the dispersion. Default to 1.

        """
        if not isinstance(observer, _TwissObserver):
            raise BeamPlottingException("The observer must be a TwissObserver.")
        if isinstance(tfs_data, cpymad.madx.Table):
            tfs_data = _pd.DataFrame(
                data={
                    "S": tfs_data.s,
                    "BETX": tfs_data.betx,
                    "BETY": tfs_data.bety,
                    "ALFX": tfs_data.alfx,
                    "ALFY": tfs_data.alfy,
                    "DX": tfs_data.dx,
                    "DY": tfs_data.dy,
                },
            )
        df_observer = observer.to_df()
        twiss_palette = kwargs.get("palette", palette)

        if with_beta:
            self._ax.plot(
                _np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values]),
                _np.hstack(
                    [
                        df_observer.iloc[0]["BETA_IN_X"],
                        df_observer.iloc[:]["BETA_OUT_X"].values,
                    ],
                ),
                "-",
                color=twiss_palette["TWISS_X"],
                linewidth=1,
                label="BETX - Manzoni",
            )

            self._ax.plot(
                _np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values]),
                _np.hstack(
                    [
                        df_observer.iloc[0]["BETA_IN_Y"],
                        df_observer.iloc[:]["BETA_OUT_Y"].values,
                    ],
                ),
                "-",
                color=twiss_palette["TWISS_Y"],
                linewidth=1,
                label="BETY - Manzoni",
            )
            if tfs_data is not None:
                self._ax.plot(
                    tfs_data["S"].values,
                    tfs_data["BETX"].values,
                    color=twiss_palette["TWISS_X"],
                    markeredgecolor=twiss_palette["TWISS_X"],
                    markersize=4,
                    marker="x",
                    ls="None",
                    label="BETX - MADX",
                )

                self._ax.plot(
                    tfs_data["S"].values,
                    tfs_data["BETY"].values,
                    color=twiss_palette["TWISS_Y"],
                    markeredgecolor=twiss_palette["TWISS_Y"],
                    markersize=4,
                    marker="x",
                    ls="None",
                    label="BETY - MADX",
                )

            self._ax.set_xlabel("S (m)")
            self._ax.set_ylabel(r"$\beta$ (m)")
            max_val = _np.ceil(_np.maximum(df_observer["BETA_OUT_X"].max(), df_observer["BETA_OUT_Y"].max()))
            self._ax.yaxis.set_major_locator(mticker.MultipleLocator(_np.ceil(max_val / 10)))
            self._ax.set_ylim([0, max_val + 5.0])

        if with_alpha:
            self._ax.plot(
                _np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values]),
                _np.hstack(
                    [
                        df_observer.iloc[0]["ALPHA_IN_X"],
                        df_observer.iloc[:]["ALPHA_OUT_X"].values,
                    ],
                ),
                "-",
                color=twiss_palette["TWISS_X"],
                linewidth=1,
                label="ALPHAX - Manzoni",
            )

            self._ax.plot(
                _np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values]),
                _np.hstack(
                    [
                        df_observer.iloc[0]["ALPHA_IN_Y"],
                        df_observer.iloc[:]["ALPHA_OUT_Y"].values,
                    ],
                ),
                "-",
                color=twiss_palette["TWISS_Y"],
                linewidth=1,
                label="ALPHAY - Manzoni",
            )
            if tfs_data is not None:
                self._ax.plot(
                    tfs_data["S"].values,
                    tfs_data["ALFX"].values,
                    color=twiss_palette["TWISS_X"],
                    markeredgecolor=twiss_palette["TWISS_X"],
                    markersize=4,
                    marker="x",
                    ls="None",
                    label="ALPHAX - MADX",
                )

                self._ax.plot(
                    tfs_data["S"].values,
                    tfs_data["ALFY"].values,
                    color=twiss_palette["TWISS_Y"],
                    markeredgecolor=twiss_palette["TWISS_Y"],
                    markersize=4,
                    marker="x",
                    ls="None",
                    label="ALPHAY - MADX",
                )

            self._ax.set_xlabel("S (m)")
            self._ax.set_ylabel(r"$\alpha$ (m)")
            max_val = _np.ceil(_np.maximum(df_observer["ALPHA_OUT_X"].max(), df_observer["ALPHA_OUT_Y"].max()))
            min_val = _np.ceil(_np.minimum(df_observer["ALPHA_OUT_X"].min(), df_observer["ALPHA_OUT_Y"].min()))
            self._ax.yaxis.set_major_locator(mticker.MultipleLocator(_np.ceil(max_val / 10)))
            self._ax.set_ylim([min_val - 5.0, max_val + 5.0])

        if with_dispersion:
            self.ax_disp = self._ax.twinx()
            self.ax_disp.plot(
                _np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values]),
                _np.hstack(
                    [
                        df_observer.iloc[0]["DISP_IN_X"],
                        df_observer.iloc[:]["DISP_OUT_X"].values,
                    ],
                ),
                "-",
                color=twiss_palette["DX"],
                linewidth=1,
                label="DX - Manzoni",
            )

            self.ax_disp.plot(
                _np.hstack([0, df_observer["AT_EXIT"].apply(lambda e: e.m_as("m")).values]),
                _np.hstack(
                    [
                        df_observer.iloc[0]["DISP_IN_Y"],
                        df_observer.iloc[:]["DISP_OUT_Y"].values,
                    ],
                ),
                "-",
                color=twiss_palette["DY"],
                linewidth=1,
                label="DY - Manzoni",
            )
            if tfs_data is not None:
                logging.warning(
                    f"Dispersion from MAD-X is multiplied by the beta relativistic factor {relativistic_beta}.",
                )
                self.ax_disp.plot(
                    tfs_data["S"].values,
                    tfs_data["DX"].values * relativistic_beta,
                    color=twiss_palette["DX"],
                    markeredgecolor=twiss_palette["DX"],
                    markersize=4,
                    marker="x",
                    ls="None",
                    label="DX - MADX",
                )

                self.ax_disp.plot(
                    tfs_data["S"].values,
                    tfs_data["DY"].values * relativistic_beta,
                    color=twiss_palette["DY"],
                    markeredgecolor=twiss_palette["DY"],
                    markersize=4,
                    marker="x",
                    ls="None",
                    label="DY - MADX",
                )

            self.ax_disp.set_xlabel("S (m)")
            self.ax_disp.set_ylabel("Dispersion (m)")
            max_val = _np.ceil(_np.maximum(df_observer["DISP_OUT_X"].max(), df_observer["DISP_OUT_Y"].max()))
            min_val = _np.floor(_np.minimum(df_observer["DISP_OUT_X"].min(), df_observer["DISP_OUT_Y"].min()))
            if max_val > 0:
                self.ax_disp.yaxis.set_major_locator(mticker.MultipleLocator(_np.ceil(max_val / 10)))
            self.ax_disp.set_ylim([min_val - 5.0, max_val + 5.0])
            self.ax_disp.legend()

    # Plotting for the phase space
    def phase_space(
        self,
        observer: _BeamObserver = None,
        element: str = None,
        location: str = "OUT",
        dim=None,
        nbins=None,
        draw_ellipse: bool = True,
    ):
        """

        Args:
            observer:
            element:
            location:
            dim:
            nbins:
            draw_ellipse:
        Returns:

        """

        if nbins is None:
            nbins = [50, 50]

        if dim is None:
            dim = ["X", "Y"]

        if not isinstance(observer, _BeamObserver):
            raise BeamPlottingException("The observer must be a BeamObserver.")

        df_observer = observer.to_df()
        if element is None:
            element = df_observer.iloc[0].name

        data_element = df_observer.loc[element, f"BEAM_{location}"]
        if location == "IN":
            s_position = df_observer.loc[element, "AT_ENTRY"]
        else:
            s_position = df_observer.loc[element, "AT_EXIT"]

        if dim[0] == "X" or dim[0] == "Y":
            unit_col_0 = "[mm]"
        else:
            unit_col_0 = "[mrad]"
        if dim[1] == "X" or dim[1] == "Y":
            unit_col_1 = "[mm]"
        else:
            unit_col_1 = "[mrad]"

        dico_plane = {"X": 0, "PX": 1, "Y": 2, "PY": 3}
        x = 1000 * data_element[:, dico_plane[dim[0]]]
        y = 1000 * data_element[:, dico_plane[dim[1]]]

        self._prepare(sposition=s_position)

        # Main Figure
        xlim = [_np.mean(x) - 5 * _np.std(x), _np.mean(x) + 5 * _np.std(x)]
        ylim = [_np.mean(y) - 5 * _np.std(y), _np.mean(y) + 5 * _np.std(y)]
        h, xedges, yedges = _np.histogram2d(x, y, bins=[nbins[0], nbins[1]], range=[xlim, ylim])
        self._fig.axes[len(self._fig.axes) - 4].imshow(
            h.T,
            extent=[xlim[0], xlim[1], ylim[0], ylim[1]],
            interpolation="nearest",
            origin="lower",
            aspect="auto",
            cmap="gist_gray_r",
        )

        self._fig.axes[len(self._fig.axes) - 4].set_xlabel(f"{dim[0]} {unit_col_0}")
        self._fig.axes[len(self._fig.axes) - 4].set_ylabel(f"{dim[1]} {unit_col_1}")
        if draw_ellipse:
            self.draw_ellipse(x, y)

        self._fig.axes[len(self._fig.axes) - 3].hist(x, bins=nbins[0], color="blue", histtype="step")
        self._fig.axes[len(self._fig.axes) - 2].hist(
            y,
            bins=nbins[1],
            orientation="horizontal",
            color="red",
            histtype="step",
        )
        bin_centerx, fitresults_x = self.histogram_fit(x, bounds_binning=nbins[0], verbose=False)
        self._fig.axes[len(self._fig.axes) - 3].plot(bin_centerx, fitresults_x.best_fit, "k--", linewidth=1)
        bin_centery, fitresults_y = self.histogram_fit(y, bounds_binning=nbins[0], verbose=False)
        self._fig.axes[len(self._fig.axes) - 2].plot(fitresults_y.best_fit, bin_centery, "k--", linewidth=1)

        # Table
        self.draw_table(x, y, dim, unit_col_0, unit_col_1, element)

    def _prepare(self, sposition):
        if len(self._fig.axes) == 1:  # We don't have a cartouche
            space = 0.02
            width, height = 0.25, 0.25
            left, hwidth = 0.05, width / 2
            bottom, hheight = 0.05, height / 2
            bottom_h = left_h = left + width + space

        else:
            bounds = self._ax.get_position().bounds
            space = 0.02
            width, height = 0.8 * (bounds[2] - bounds[0]) - space, 0.8 * (bounds[3] - bounds[1]) - space
            left, hwidth = bounds[0], bounds[2] - width - space
            bottom, hheight = bounds[1], bounds[3] - height - space
            bottom_h = left + height + space
            left_h = left + width + space
            self._ax2.add_patch(
                patches.Rectangle(
                    (sposition.m_as("m"), 1.15 - 0.075),
                    0.01,
                    0.15,
                    hatch="",
                    facecolor="m",
                    clip_on=False,
                ),
            )
            self._ax2.spines["top"].set_visible(False)
            self._ax2.spines["bottom"].set_visible(False)
            self._ax2.spines["right"].set_visible(False)
            self._ax2.spines["left"].set_visible(False)

        self._fig.delaxes(self._ax)
        # Set up the geometry of the three plots
        rect_beam = [left, bottom, width, height]  # dimensions of temp plot
        rect_histx = [left, bottom_h, width, hheight]  # dimensions of x-histogram
        rect_histy = [left_h, bottom, hwidth, height]  # dimensions of y-histogram
        rect_tab = [left_h, bottom_h, hwidth, hheight]  # dimensions of tab
        # Make the three plots
        _ = self._fig.add_axes(rect_beam)  # noqa: F841
        ax_histx = self._fig.add_axes(rect_histx)  # x histogram
        ax_histy = self._fig.add_axes(rect_histy)  # y histogram
        ax_histx.set_ylabel("Counts")
        ax_histx.grid(True)
        ax_histy.set_xlabel("Counts")
        ax_histy.grid(True)
        nullfmt = mticker.NullFormatter()
        ax_histx.xaxis.set_major_formatter(nullfmt)
        ax_histy.yaxis.set_major_formatter(nullfmt)
        ax_tab = self._fig.add_axes(rect_tab)  # y histogram
        ax_tab.tick_params(labelbottom="off", labelleft="off", left="off", bottom="off")

    def draw_ellipse(self, x, y):
        [e_val, evec] = _np.linalg.eig(_np.cov(x, y))
        chisq = [2.278868566, 5.991464547, 11.61828598]
        x_r, y_r = [], []
        for i in range(0, len(chisq)):
            x_r.append(2 * (chisq[i] * e_val[0]) ** 0.5)
            y_r.append(2 * (chisq[i] * e_val[1]) ** 0.5)

        ang = self.rotation_angle(e_val, evec)

        self._fig.axes[len(self._fig.axes) - 4].add_patch(
            self.ellipse(
                x_r[0],
                y_r[0],
                ang,
                _np.mean(x),
                _np.mean(y),
                color="red",
                label="$1\\sigma$",
            ),
        )
        self._fig.axes[len(self._fig.axes) - 4].add_patch(
            self.ellipse(
                x_r[1],
                y_r[1],
                ang,
                _np.mean(x),
                _np.mean(y),
                color="blue",
                label="$2\\sigma$",
            ),
        )
        self._fig.axes[len(self._fig.axes) - 4].add_patch(
            self.ellipse(
                x_r[2],
                y_r[2],
                ang,
                _np.mean(x),
                _np.mean(y),
                color="green",
                label="$3\\sigma$",
            ),
        )
        self._fig.axes[len(self._fig.axes) - 4].legend()

    def draw_table(self, x, y, dim, unit_col_0, unit_col_1, element):
        mean_x = f"{_np.round(_np.mean(x), 3)}"
        std_x = f"{_np.round(_np.std(x), 3)}"
        median_x = f"{_np.round(_np.median(x), 3)}"
        mean_y = f"{_np.round(_np.mean(y), 3)}"
        std_y = f"{_np.round(_np.std(y), 3)}"
        median_y = f"{_np.round(_np.median(y), 3)}"

        self._fig.axes[-1].tick_params(labelbottom="off", labelleft="off", left="off", bottom="off")
        self._fig.axes[-1].get_xaxis().set_visible(False)
        self._fig.axes[-1].get_yaxis().set_visible(False)
        x0 = self._fig.axes[-1].get_xlim()[0]
        y0 = self._fig.axes[-1].get_ylim()[0]
        self._fig.axes[-1].axvline(x0 + 0.25, color="k", linewidth=1)
        self._fig.axes[-1].axvline(x0 + 0.5, color="k", linewidth=1)
        self._fig.axes[-1].axvline(x0 + 0.75, color="k", linewidth=1)
        self._fig.axes[-1].axhline(y0 + 0.4, color="k", linewidth=1)
        self._fig.axes[-1].axhline(y0 + 0.8, color="k", linewidth=1)
        self._fig.axes[-1].axhline(y0 + 0.83, color="k", linewidth=1)
        self._fig.axes[-1].annotate(
            element,
            xy=(x0 + 0.125, y0 + 0.9),
            xytext=(x0 + 0.125, y0 + 0.9),
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=12,
        )
        self._fig.axes[-1].annotate(
            "mean",
            xy=(x0 + 0.375, y0 + 0.9),
            xytext=(x0 + 0.375, y0 + 0.9),
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=12,
        )
        self._fig.axes[-1].annotate(
            "std ",
            xy=(x0 + 0.625, y0 + 0.9),
            xytext=(x0 + 0.625, y0 + 0.9),
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=12,
        )
        self._fig.axes[-1].annotate(
            "median",
            xy=(x0 + 0.875, y0 + 0.9),
            xytext=(x0 + 0.875, y0 + 0.9),
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=12,
        )

        xname = f"{dim[0]} {unit_col_0}"
        yname = f"{dim[1]} {unit_col_1}"
        self._fig.axes[-1].annotate(
            xname,
            xy=(x0 + 0.125, 0.6),
            xytext=(x0 + 0.125, 0.6),
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=11,
        )
        self._fig.axes[-1].annotate(
            yname,
            xy=(x0 + 0.125, 0.2),
            xytext=(x0 + 0.125, 0.2),
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=12,
        )
        self._fig.axes[-1].annotate(
            mean_x,
            xy=(x0 + 0.375, 0.6),
            xytext=(x0 + 0.375, 0.6),
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=12,
        )
        self._fig.axes[-1].annotate(
            std_x,
            xy=(x0 + 0.625, 0.6),
            xytext=(x0 + 0.625, 0.6),
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=12,
        )
        self._fig.axes[-1].annotate(
            median_x,
            xy=(x0 + 0.875, 0.6),
            xytext=(x0 + 0.875, 0.6),
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=12,
        )
        self._fig.axes[-1].annotate(
            mean_y,
            xy=(x0 + 0.375, 0.2),
            xytext=(x0 + 0.375, 0.2),
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=12,
        )
        self._fig.axes[-1].annotate(
            std_y,
            xy=(x0 + 0.625, 0.2),
            xytext=(x0 + 0.625, 0.2),
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=12,
        )
        self._fig.axes[-1].annotate(
            median_y,
            xy=(x0 + 0.875, 0.2),
            xytext=(x0 + 0.875, 0.2),
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=12,
        )

    # Plotting five spot map
    def five_spot_map(self, bl_track0, bl_track1, bl_track2, bl_track3, bl_track4):
        # TODO
        pass
        # try:
        #     # Order the 5 simulations outputs
        #     center = _pd.DataFrame(bl_track0.line['BEAM']['ISO'].distribution['X'])
        #     blcorner = _pd.DataFrame(bl_track1.line['BEAM']['ISO'].distribution['X'])
        #     brcorner = _pd.DataFrame(bl_track2.line['BEAM']['ISO'].distribution['X'])
        #     tlcorner = _pd.DataFrame(bl_track3.line['BEAM']['ISO'].distribution['X'])
        #     trcorner = _pd.DataFrame(bl_track4.line['BEAM']['ISO'].distribution['X'])
        #     datax = _pd.concat([center, blcorner, brcorner, tlcorner, trcorner], ignore_index=True)
        #
        #     center = _pd.DataFrame(bl_track0.line['BEAM']['ISO'].distribution['Y'])
        #     blcorner = _pd.DataFrame(bl_track1.line['BEAM']['ISO'].distribution['Y'])
        #     brcorner = _pd.DataFrame(bl_track2.line['BEAM']['ISO'].distribution['Y'])
        #     tlcorner = _pd.DataFrame(bl_track3.line['BEAM']['ISO'].distribution['Y'])
        #     trcorner = _pd.DataFrame(bl_track4.line['BEAM']['ISO'].distribution['Y'])
        #     datay = _pd.concat([center, blcorner, brcorner, tlcorner, trcorner], ignore_index=True)
        # except:
        #     print('Error, one simulation is missing. You have to put the 5 simulations outputs in the functions')
        #
        # # Create the 2D histogram with the 5 spots.
        # _ = plt.hist2d(datax['X'].values, datay['Y'].values, bins=400, cmap='gist_gray_r')
        # data_X = []
        # data_X.append(1e3 * bl_track0.line['BEAM']['ISO'].std['X'])
        # data_X.append(1e3 * bl_track1.line['BEAM']['ISO'].std['X'])
        # data_X.append(1e3 * bl_track2.line['BEAM']['ISO'].std['X'])
        # data_X.append(1e3 * bl_track3.line['BEAM']['ISO'].std['X'])
        # data_X.append(1e3 * bl_track4.line['BEAM']['ISO'].std['X'])
        #
        # data_S = []
        # data_S.append(100 * _np.abs(bl_track0.line['BEAM']['ISO'].std['X'] -
        # bl_track0.line['BEAM']['ISO'].std['Y']) / (
        #         bl_track0.line['BEAM']['ISO'].std['X'] + bl_track0.line['BEAM']['ISO'].std['Y']))
        # data_S.append(100 * _np.abs(bl_track1.line['BEAM']['ISO'].std['X'] - bl_track1.line['BEAM']['ISO']
        # .std['Y']) / (
        #         bl_track1.line['BEAM']['ISO'].std['X'] + bl_track1.line['BEAM']['ISO'].std['Y']))
        # data_S.append(100 * _np.abs(bl_track2.line['BEAM']['ISO'].std['X'] - bl_track2.line['BEAM']['ISO']
        # .std['Y']) / (
        #         bl_track2.line['BEAM']['ISO'].std['X'] + bl_track2.line['BEAM']['ISO'].std['Y']))
        # data_S.append(100 * _np.abs(bl_track3.line['BEAM']['ISO'].std['X'] - bl_track3.line['BEAM']['ISO']
        # .std['Y']) / (
        #         bl_track3.line['BEAM']['ISO'].std['X'] + bl_track3.line['BEAM']['ISO'].std['Y']))
        # data_S.append(100 * _np.abs(bl_track4.line['BEAM']['ISO'].std['X'] - bl_track4.line['BEAM']['ISO']
        # .std['Y']) / (
        #         bl_track4.line['BEAM']['ISO'].std['X'] + bl_track4.line['BEAM']['ISO'].std['Y']))
        #
        # data_T = []
        # data_T.append(_np.degrees(
        #     self.ellipse_angle_of_rotation(self.fitEllipse(1e3 * bl_track0.line['BEAM']['ISO'].distribution['X'],
        #                                                    1e3 * bl_track0.line['BEAM']['ISO'].distribution[
        #                                                        'Y']))))
        # data_T.append(_np.degrees(
        #     self.ellipse_angle_of_rotation(self.fitEllipse(1e3 * bl_track1.line['BEAM']['ISO'].distribution['X'],
        #                                                    1e3 * bl_track1.line['BEAM']['ISO'].distribution[
        #                                                        'Y']))))
        # data_T.append(_np.degrees(
        #     self.ellipse_angle_of_rotation(self.fitEllipse(1e3 * bl_track2.line['BEAM']['ISO'].distribution['X'],
        #                                                    1e3 * bl_track2.line['BEAM']['ISO'].distribution[
        #                                                        'Y']))))
        # data_T.append(_np.degrees(
        #     self.ellipse_angle_of_rotation(self.fitEllipse(1e3 * bl_track3.line['BEAM']['ISO'].distribution['X'],
        #                                                    1e3 * bl_track3.line['BEAM']['ISO'].distribution[
        #                                                        'Y']))))
        # data_T.append(_np.degrees(
        #     self.ellipse_angle_of_rotation(self.fitEllipse(1e3 * bl_track4.line['BEAM']['ISO'].distribution['X'],
        #                                                    1e3 * bl_track4.line['BEAM']['ISO'].distribution[
        #                                                        'Y']))))
        #
        # plt.annotate(
        #     rf"$\sigma_X = {round(data_X[1], 2)}$mm \n $S = {round(data_S[1], 2)}$% \n $\phi={round(data_T[1], 2)} °$"
        #     , xy=(- 0.1 + 0.03, 0.1), xytext=(- 0.1 + 0.03, 0.1),
        #     horizontalalignment='center', verticalalignment='center',
        #     fontsize=12)
        # plt.annotate(
        #     rf"$\sigma_X = {round(data_X[4], 2)}$mm \n $S = {round(data_S[4], 2)}$% \n $\phi={round(data_T[4], 2)} °$"
        #     , xy=(0.1 - 0.03, 0.1), xytext=(0.1 - 0.03, 0.1),
        #     horizontalalignment='center', verticalalignment='center',
        #     fontsize=12)
        # plt.annotate(
        #     rf"$\sigma_X = {round(data_X[2], 2)}$mm \n $S = {round(data_S[2], 2)}$% \n $\phi={round(data_T[2], 2)} °$"
        #     , xy=(- 0.1 + 0.03, -0.1), xytext=(- 0.1 + 0.03, -0.1),
        #     horizontalalignment='center', verticalalignment='center',
        #     fontsize=12)
        # plt.annotate(
        #     rf"$\sigma_X = {round(data_X[3], 2)}$mm \n $S = {round(data_S[3], 2)}$% \n $\phi={round(data_T[3], 2)} °$"
        #     , xy=(0.1 - 0.03, -0.1), xytext=(0.1 - 0.03, -0.1),
        #     horizontalalignment='center', verticalalignment='center',
        #     fontsize=12)
        # plt.annotate(
        #     rf"$\sigma_X = {round(data_X[0], 2)}$mm \n $S = {round(data_S[0], 2)}$% \n $\phi={round(data_T[0], 2)} °$"
        #     , xy=(0.0 + 0.03, 0.0), xytext=(0.0 + 0.03, 0.0),
        #     horizontalalignment='center', verticalalignment='center',
        #     fontsize=12)



import numba as _nb
import numpy as _np
from numba import njit


@njit(parallel=True)
def drift2(b1: _np.ndarray, b2: _np.ndarray, length: float):
    for i in _nb.prange(0, b1.shape[0]):
        b2[i, 0] = length * b1[i, 1]


@njit(parallel=True)
def drift4(b1: _np.ndarray, b2: _np.ndarray, kargs: _np.ndarray):
    """
    Performance tests on MBP i9 show that the explicit loop is faster, both at compile-time and at runtime
    than the nb broadcast variant.

    Args:
        b1: a numpy array containing all the particles
        b2: explicit destination for the result
        kargs: the drift length (in meters)

    Returns:
        the destination (result) array
    """
    for i in _nb.prange(0, b1.shape[0]):
        b2[i, 0] = b1[i, 0] + kargs[0] * b1[i, 1]
        b2[i, 1] = b1[i, 1]
        b2[i, 2] = b1[i, 2] + kargs[0] * b1[i, 3]
        b2[i, 3] = b1[i, 3]
    return b2


@njit(parallel=True)
def drift5(b1: _np.ndarray, b2: _np.ndarray, kargs: _np.ndarray):
    """
    Performance tests on MBP i9 show that the explicit loop is faster, both at compile-time and at runtime
    than the nb broadcast variant.

    Args:
        b1: a numpy array containing all the particles
        b2: explicit destination for the result
        kargs: the drift length (in meters)

    Returns:
        the destination (result) array
    """
    for i in _nb.prange(0, b1.shape[0]):
        b2[i, 0] = b1[i, 0] + kargs[0] * b1[i, 1]
        b2[i, 1] = b1[i, 1]
        b2[i, 2] = b1[i, 2] + kargs[0] * b1[i, 3]
        b2[i, 3] = b1[i, 3]
        b2[i, 4] = b1[i, 4]
    return b2


@njit(parallel=True)
def drift6(b1: _np.ndarray, b2: _np.ndarray, length: float):
    """

    .. note:: Performance tests on MBP i9 show that the explicit loop is faster, both at compile-time and at runtime
    than the nb broadcast variant.

    Args:
        b1: a numpy array containing all the particles
        b2: explicit destination for the result
        length: the drift length (in meters)

    Returns:
        the destination (result) array
    """
    for i in _nb.prange(0, b1.shape[0]):
        b2[i, 0] = b1[i, 0] + length * b1[i, 1]
        b2[i, 1] = b1[i, 1]
        b2[i, 2] = b1[i, 2] + length * b1[i, 3]
        b2[i, 3] = b1[i, 3]
        b2[i, 4] = b1[i, 4]
        b2[i, 5] = b1[i, 5]
    return b1, b2


@njit(parallel=True, fastmath=True)
def compute_mad_drift_matrix(element_parameters: list, **_) -> _np.ndarray:
    L: float = element_parameters[0]
    R = _np.zeros((6, 6))

    # Definition of the matrix elements
    R[0, 0] = 1
    R[0, 1] = L
    R[1, 1] = 1
    R[2, 2] = 1
    R[2, 3] = L
    R[3, 3] = 1
    R[4, 4] = 1
    R[5, 5] = 1

    return R


@njit(parallel=True, fastmath=True)
def compute_mad_drift_tensor(element_parameters: list, **_) -> _np.ndarray:
    T = _np.zeros((6, 6, 6))

    return T



import numpy as np
from numba import njit
from numba.typed import List as nList
from numpy import cos, cosh, sin, sinh, sqrt


@njit(cache=True)
def compute_transport_quadrupole_ex_matrix(element_parameters: nList) -> np.ndarray:
    L: float = element_parameters[0]
    k1: float = element_parameters[1]
    d: float = element_parameters[2]
    k1 = k1 / (1 + d)
    R = np.zeros((6, 6))
    R[4, 4] = 1
    R[5, 5] = 1
    if k1 == 0:
        R[0, 0] = 1
        R[0, 1] = L
        R[1, 1] = 1
        R[2, 2] = 1
        R[2, 3] = L
        R[3, 3] = 1
    if k1 > 0:
        R[0, 0] = cos(L * sqrt(k1))
        R[0, 1] = sin(L * sqrt(k1)) / sqrt(k1)
        R[1, 0] = -sqrt(k1) * sin(L * sqrt(k1))
        R[1, 1] = cos(L * sqrt(k1))
        R[2, 2] = cosh(L * sqrt(k1))
        R[2, 3] = sinh(L * sqrt(k1)) / sqrt(k1)
        R[3, 2] = sqrt(k1) * sinh(L * sqrt(k1))
        R[3, 3] = cosh(L * sqrt(k1))
    if k1 < 0:
        R[0, 0] = cosh(L * sqrt(-k1))
        R[0, 1] = sinh(L * sqrt(-k1)) / sqrt(-k1)
        R[1, 0] = sqrt(-k1) * sinh(L * sqrt(-k1))
        R[1, 1] = cosh(L * sqrt(-k1))
        R[2, 2] = cos(L * sqrt(-k1))
        R[2, 3] = sin(L * sqrt(-k1)) / sqrt(-k1)
        R[3, 2] = -sqrt(-k1) * sin(L * sqrt(-k1))
        R[3, 3] = cos(L * sqrt(-k1))
    return R


@njit(cache=True)
def compute_transport_quadrupole_ex_tensor(element_parameters: nList) -> np.ndarray:
    L: float = element_parameters[0]
    k1: float = element_parameters[1]
    d: float = element_parameters[len(element_parameters) - 1]
    k1 = k1 / (1 + d)
    T = np.zeros((6, 6, 6))
    if k1 > 0:
        T[0, 0, 5] = L * sqrt(k1) * sin(L * sqrt(k1)) / 2
        T[0, 1, 5] = -L * cos(L * sqrt(k1)) / 2 + sin(L * sqrt(k1)) / (2 * sqrt(k1))
        T[1, 0, 5] = L * k1 * cos(L * sqrt(k1)) / 2 + sqrt(k1) * sin(L * sqrt(k1)) / 2
        T[1, 1, 5] = L * sqrt(k1) * sin(L * sqrt(k1)) / 2
        T[2, 2, 5] = -L * sqrt(k1) * sinh(L * sqrt(k1)) / 2
        T[2, 3, 5] = -L * cosh(L * sqrt(k1)) / 2 + sinh(L * sqrt(k1)) / (2 * sqrt(k1))
        T[3, 2, 5] = -L * k1 * cosh(L * sqrt(k1)) / 2 - sqrt(k1) * sinh(L * sqrt(k1)) / 2
        T[3, 3, 5] = -L * sqrt(k1) * sinh(L * sqrt(k1)) / 2
    if k1 < 0:
        T[0, 0, 5] = -L * sqrt(-k1) * sinh(L * sqrt(-k1)) / 2
        T[0, 1, 5] = -L * cosh(L * sqrt(-k1)) / 2 + sinh(L * sqrt(-k1)) / (2 * sqrt(-k1))
        T[1, 0, 5] = L * k1 * cosh(L * sqrt(-k1)) / 2 - sqrt(-k1) * sinh(L * sqrt(-k1)) / 2
        T[1, 1, 5] = -L * sqrt(-k1) * sinh(L * sqrt(-k1)) / 2
        T[2, 2, 5] = L * sqrt(-k1) * sin(L * sqrt(-k1)) / 2
        T[2, 3, 5] = -L * cos(L * sqrt(-k1)) / 2 + sin(L * sqrt(-k1)) / (2 * sqrt(-k1))
        T[3, 2, 5] = -L * k1 * cos(L * sqrt(-k1)) / 2 + sqrt(-k1) * sin(L * sqrt(-k1)) / 2
        T[3, 3, 5] = L * sqrt(-k1) * sin(L * sqrt(-k1)) / 2
    return T



"""
TODO
"""
from typing import Optional, Tuple

import numpy as _np

from ... import ureg as _ureg
from ..integrators import MadXIntegrator, MadXParaxialDriftIntegrator
from ..kernels import batched_vector_matrix
from .elements import ManzoniElement as _ManzoniElement


class Marker(_ManzoniElement):
    """
    Define a Marker
    """

    INTEGRATOR = None

    def propagate(self, beam_in: _np.ndarray, beam_out: _np.ndarray = None, parameters: list = None):
        _np.copyto(dst=beam_out, src=beam_in, casting="no")
        return beam_in, beam_out


class Matrix(_ManzoniElement):
    """
    Define a Matrix element.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the Matrix with their default values.

    Examples:
        >>> m1 = Matrix('M1', MATRIX=_np.eye(6))
        >>> m1 #doctest: +NORMALIZE_WHITESPACE
            Matrix: {'NAME': 'M1',
            'AT_ENTRY': <Quantity(0, 'meter')>,
            'AT_CENTER': <Quantity(0, 'meter')>,
            'AT_EXIT': <Quantity(0, 'meter')>,
            'MATRIX': array([[1., 0., 0., 0., 0., 0.],
                             [0., 1., 0., 0., 0., 0.],
                             [0., 0., 1., 0., 0., 0.],
                             [0., 0., 0., 1., 0., 0.],
                             [0., 0., 0., 0., 1., 0.],
                             [0., 0., 0., 0., 0., 1.]])}
    """

    INTEGRATOR = None
    PARAMETERS = {
        "MATRIX": (_np.eye(6), "Transfer matrix."),
    }

    def propagate(
        self,
        beam_in: _np.ndarray,
        beam_out: Optional[_np.ndarray] = None,
        global_parameters: list = None,
    ) -> Tuple[_np.ndarray, _np.ndarray]:
        if self.integrator not in [MadXIntegrator, MadXParaxialDriftIntegrator]:
            beam_in[:, [-2, -1]] = beam_in[:, [-1, -2]]

        bvm = batched_vector_matrix(beam_in, beam_out, self.MATRIX)

        if self.integrator not in [MadXIntegrator, MadXParaxialDriftIntegrator]:
            bvm[0][:, [-1, -2]] = bvm[0][:, [-2, -1]]
            bvm[1][:, [-1, -2]] = bvm[1][:, [-2, -1]]

        return bvm

    @property
    def parameters(self) -> list:
        return [
            self.MATRIX,
        ]


class Gap(_ManzoniElement):
    """
    Define a gap where no physical geometry is placed

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the Gap with their default values.

    Examples:
        >>> g1 = Gap('G1', L=1*_ureg.m)
        >>> g1 #doctest: +NORMALIZE_WHITESPACE
            Gap: {'NAME': 'G1',
            'AT_ENTRY': <Quantity(0, 'meter')>,
            'AT_CENTER': <Quantity(0, 'meter')>,
            'AT_EXIT': <Quantity(0, 'meter')>,
            'L': <Quantity(1, 'meter')>,
            'APERTYPE': None,
            'APERTURE': []}
    """

    PARAMETERS = {
        "L": (0.0 * _ureg.m, "Drift length."),
        "APERTYPE": (None, "Aperture type (CIRCULAR, ELLIPTICAL, RECTANGULAR or PHASE_SPACE)"),
        "APERTURE": ([], ""),
    }
    """Parameters of the element, with their default value and their descriptions."""

    def propagate(
        self,
        beam_in: _np.ndarray,
        beam_out: _np.ndarray = None,
        global_parameters: list = None,
    ) -> Tuple[_np.ndarray, _np.ndarray]:
        if self.L.magnitude == 0:
            _np.copyto(dst=beam_out, src=beam_in, casting="no")
            return beam_in, beam_out
        else:
            return self.integrator.propagate(self, beam_in, beam_out, global_parameters)

    @property
    def parameters(self) -> list:
        return list(
            map(
                float,
                [
                    self.L.m_as("meter"),
                ],
            ),
        )


class Drift(Gap):
    """
    Definition of a Drift.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the Drift with their default values.

    Examples:
        >>> d1 = Drift('D1', L=1*_ureg.m, APERTYPE='CIRCULAR', APERTURE=[5*_ureg.cm])
        >>> d1 #doctest: +NORMALIZE_WHITESPACE
            Drift: {'NAME': 'D1',
                    'AT_ENTRY': <Quantity(0, 'meter')>,
                    'AT_CENTER': <Quantity(0, 'meter')>,
                    'AT_EXIT': <Quantity(0, 'meter')>,
                    'L': <Quantity(1, 'meter')>,
                    'APERTYPE': 'CIRCULAR',
                    'APERTURE': [<Quantity(5, 'centimeter')>]}
    """

    PARAMETERS = {
        "APERTYPE": (None, "Aperture type (CIRCULAR, ELLIPTICAL, RECTANGULAR or PHASE_SPACE)"),
        "APERTURE": ([], ""),
    }


class SRotation(_ManzoniElement):
    """
    Definition of a SRotation.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the SRotation with their default values.

    Examples:
        >>> s1 = SRotation('S1', ANGLE=10*_ureg.degrees)
        >>> s1 #doctest: +NORMALIZE_WHITESPACE
            SRotation: {'NAME': 'S1',
                        'AT_ENTRY': <Quantity(0, 'meter')>,
                        'AT_CENTER': <Quantity(0, 'meter')>,
                        'AT_EXIT': <Quantity(0, 'meter')>,
                        'ANGLE': <Quantity(10, 'degree')>}
    """

    PARAMETERS = {
        "ANGLE": (0.0 * _ureg.radian, "Angle of rotation along the s-axis."),
    }

    @property
    def parameters(self) -> list:
        return list(
            map(
                float,
                [
                    self.ANGLE.m_as("radian"),
                ],
            ),
        )


class Magnet(_ManzoniElement):
    PARAMETERS = {
        "APERTYPE": (None, "Aperture type (CIRCULAR, ELLIPTICAL, RECTANGULAR or PHASE_SPACE)"),
        "APERTURE": ([], ""),
        "KINEMATICS": (None, "Reference kinematics"),
    }


class Quadrupole(Magnet):
    """
    Define a Quadrupole magnet.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the Quadrupole with their default values.

    Examples:
        >>> q1 = Quadrupole('Q1', L=1*_ureg.m, K1=3*_ureg.m**-2)
        >>> q1 #doctest: +NORMALIZE_WHITESPACE
            Quadrupole: {'NAME': 'Q1',
            'AT_ENTRY': <Quantity(0, 'meter')>,
            'AT_CENTER': <Quantity(0, 'meter')>,
            'AT_EXIT': <Quantity(0, 'meter')>,
            'APERTYPE': None, 'APERTURE': [],
            'KINEMATICS': None,
            'L': <Quantity(1, 'meter')>,
            'K1': <Quantity(3, '1 / meter ** 2')>,
            'K1S': <Quantity(0.0, '1 / meter ** 2')>,
            'TILT': <Quantity(0.0, 'radian')>}
    """

    PARAMETERS = {
        "L": (0.0 * _ureg.m, "Quadrupole length."),
        "K1": (0.0 * _ureg.m**-2, "Normalized gradient."),
        "K1S": (0.0 * _ureg.m**-2, "Normalized skew gradient."),
        "TILT": (0.0 * _ureg.radian, "Magnet tilt angle."),
    }

    @property
    def parameters(self) -> list:
        k1 = self.K1.m_as("m**-2")
        k1s = self.K1S.m_as("m**-2")
        tilt = self.TILT.m_as("radian")

        if k1s != 0.0 or tilt != 0.0:
            tilt -= _np.arctan2(k1s, k1) / 2
            k1 = _np.sqrt(k1**2 + k1s**2)

        return list(
            map(
                float,
                [
                    self.L.m_as("m"),
                    k1,
                    tilt,
                ],
            ),
        )


class Bend(Magnet):
    """
    Define a Bend magnet.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the Bend with their default values.

    Examples:
        >>> b1 = Bend('B1', L=1*_ureg.m, ANGLE=30*_ureg.degrees, E1=5*_ureg.degrees, HGAP=2*_ureg.cm, K1=3*_ureg.m**-2)
        >>> b1 #doctest: +NORMALIZE_WHITESPACE
            Bend: {'NAME': 'B1',
            'AT_ENTRY': <Quantity(0, 'meter')>,
            'AT_CENTER': <Quantity(0, 'meter')>,
            'AT_EXIT': <Quantity(0, 'meter')>,
            'APERTYPE': None, 'APERTURE': [],
            'KINEMATICS': None,
            'ANGLE': <Quantity(30, 'degree')>,
            'K0': <Quantity(0.0, '1 / meter')>,
            'K1': <Quantity(3, '1 / meter ** 2')>,
            'K2': <Quantity(0.0, '1 / meter ** 3')>,
            'L': <Quantity(1, 'meter')>,
            'E1': <Quantity(5, 'degree')>,
            'E2': <Quantity(0.0, 'radian')>,
            'TILT': <Quantity(0.0, 'radian')>,
            'HGAP': <Quantity(2, 'centimeter')>,
            'FINT': 0.0,
            'FINTX': 0.0}
    """

    PARAMETERS = {
        "ANGLE": (0.0 * _ureg.radian, "Bending angle."),
        "K0": (0.0 * _ureg.m**-1, "Dipolar normalized gradient"),
        "K1": (0.0 * _ureg.m**-2, "Quadrupolar normalized gradient."),
        "K2": (0.0 * _ureg.m**-3, "Sextupolar normalized gradient."),
        "L": (0.0 * _ureg.m, "Magnet length."),
        "E1": (0.0 * _ureg.radian, "Entrance face angle."),
        "E2": (0.0 * _ureg.radian, "Exit face angle."),
        "TILT": (0.0 * _ureg.radian, "Magnet tilt angle."),
        "HGAP": (0.0 * _ureg.m, "Magnet gap."),
        "FINT": (0.0, "Fringe field integral."),
        "FINTX": (0.0, "Exit fringe field integral."),
    }

    @staticmethod
    def compute_fringe(h: float, e: float, hgap: float, fint: float) -> Tuple[float, float]:
        fringe_x = h * _np.tan(e)
        corr = h * (2 * hgap) * fint
        psi = e - corr * (1.0 / _np.cos(e)) * (1 + _np.sin(e) ** 2)
        fringe_y = -h * _np.tan(psi)
        return fringe_x, fringe_y

    @property
    def length(self) -> float:
        return self.L.m_as("m")

    @property
    def edges(self) -> Tuple[float, float]:
        return self.E1.m_as("radian"), self.E2.m_as("radian")

    @property
    def fringe_field_integrals(self) -> Tuple[float, float]:
        fint = self.FINT
        fintx = self.FINTX if self.FINTX >= 0 else fint  # For exact compatibility with MAD-X
        return fint, fintx

    @property
    def parameters(self) -> list:
        # Generic parameters
        length = self.length
        h = self.ANGLE.m_as("radian") / self.length
        if self.K0.m_as("m**-1") is None or self.K0.m_as("m**-1") == 0:
            k0 = h
        else:
            k0 = self.K0.m_as("m**-1")
        hgap = self.HGAP.m_as("m")
        e1, e2 = self.edges
        fint, fintx = self.fringe_field_integrals
        entrance_fringe_x, entrance_fringe_y = Bend.compute_fringe(h, e1, hgap, fint)
        exit_fringe_x, exit_fringe_y = Bend.compute_fringe(h, e2, hgap, fintx)

        return list(
            map(
                float,
                [
                    length,  # 0
                    self.ANGLE.m_as("radian"),  # 1
                    self.K1.m_as("m**-2"),  # 2
                    self.K2.m_as("m**-3"),  # 3
                    self.TILT.m_as("radian"),  # 4
                    h,  # 5
                    k0,  # 6
                    entrance_fringe_x,  # 7
                    entrance_fringe_y,  # 8
                    exit_fringe_x,  # 9
                    exit_fringe_y,  # 10
                ],
            ),
        )


class SBend(Bend):
    """
    Definition of a SBend
    """

    pass


class RBend(Bend):
    """
    Definition of a RBend
    """

    @property
    def length(self) -> float:
        length = self.L.m_as("m")
        angle = self.ANGLE.m_as("rad")
        if angle > 1e-8:
            return length * angle / (2.0 * _np.sin(angle / 2.0))
        else:
            return length

    @property
    def edges(self) -> Tuple[float, float]:
        alpha = self.ANGLE.m_as("radian") / 2.0
        return self.E1.m_as("radian") + alpha, self.E2.m_as("radian") + alpha


class Fringein(Bend):
    """
    Define a Fringe field at the entrance of a bending magnet.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the Fringein with their default values.

    Examples:
        >>> f1 = Fringein('F1', L=1*_ureg.mm, ANGLE=30*_ureg.degrees, E1=5*_ureg.degrees,
        ...               HGAP=2*_ureg.cm, K1=3*_ureg.m**-2, R1=1*_ureg.cm)
        >>> f1 #doctest: +NORMALIZE_WHITESPACE
            Fringein: {'NAME': 'F1',
                       'AT_ENTRY': <Quantity(0, 'meter')>,
                       'AT_CENTER': <Quantity(0, 'meter')>,
                       'AT_EXIT': <Quantity(0, 'meter')>,
                       'APERTYPE': None, 'APERTURE': [],
                       'KINEMATICS': None,
                       'ANGLE': <Quantity(30, 'degree')>,
                       'K0': <Quantity(0.0, '1 / meter')>,
                       'K1': <Quantity(3, '1 / meter ** 2')>,
                       'K2': <Quantity(0.0, '1 / meter ** 3')>,
                       'L': <Quantity(1, 'millimeter')>,
                       'E1': <Quantity(5, 'degree')>,
                       'E2': <Quantity(0.0, 'radian')>,
                       'TILT': <Quantity(0.0, 'radian')>,
                       'HGAP': <Quantity(2, 'centimeter')>,
                       'FINT': 0.0, 'FINTX': 0.0,
                       'R1': <Quantity(1, 'centimeter')>}
    """

    PARAMETERS = {
        "ANGLE": (0.0 * _ureg.radian, "Bending angle."),
        "L": (0.0 * _ureg.m, "Magnet length."),
        "K1": (0.0 * _ureg.m**-2, "Quadrupolar normalized gradient."),
        "E1": (0.0 * _ureg.radian, "Entrance face angle."),
        "HGAP": (0.0 * _ureg.m, "Magnet gap."),
        "FINT": (0.0, "Fringe field integral."),
        "R1": (_np.inf * _ureg.m, "Entrance face curvature radius"),
    }

    @property
    def parameters(self) -> list:
        # Generic parameters
        return list(
            map(
                float,
                [
                    self.length,  # 0
                    self.ANGLE.m_as("radian"),  # 1
                    self.K1.m_as("m**-2"),  # 2
                    self.E1.m_as("radian"),  # 3
                    self.HGAP.m_as("m"),  # 4
                    self.FINT,  # 5
                    self.R1.m_as("m"),  # 6
                ],
            ),
        )


class Fringeout(Bend):
    """
    Define a Fringe field at the exit of a bending magnet.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the Fringeout with their default values.

    Examples:
        >>> f2 = Fringeout('F2', L=1*_ureg.mm, ANGLE=30*_ureg.degrees, E1=5*_ureg.degrees,
        ...                 HGAP=2*_ureg.cm, K1=3*_ureg.m**-2, R2=1*_ureg.cm)
        >>> f2 #doctest: +NORMALIZE_WHITESPACE
            Fringeout: {'NAME': 'F2',
                       'AT_ENTRY': <Quantity(0, 'meter')>,
                       'AT_CENTER': <Quantity(0, 'meter')>,
                       'AT_EXIT': <Quantity(0, 'meter')>,
                       'APERTYPE': None, 'APERTURE': [],
                       'KINEMATICS': None,
                       'ANGLE': <Quantity(30, 'degree')>,
                       'K0': <Quantity(0.0, '1 / meter')>,
                       'K1': <Quantity(3, '1 / meter ** 2')>,
                       'K2': <Quantity(0.0, '1 / meter ** 3')>,
                       'L': <Quantity(1, 'millimeter')>,
                       'E1': <Quantity(5, 'degree')>,
                       'E2': <Quantity(0.0, 'radian')>,
                       'TILT': <Quantity(0.0, 'radian')>,
                       'HGAP': <Quantity(2, 'centimeter')>,
                       'FINT': 0.0, 'FINTX': 0.0,
                       'R2': <Quantity(1, 'centimeter')>}
    """

    PARAMETERS = {
        "ANGLE": (0.0 * _ureg.radian, "Bending angle."),
        "L": (0.0 * _ureg.m, "Magnet length."),
        "K1": (0.0 * _ureg.m**-2, "Quadrupolar normalized gradient."),
        "E2": (0.0 * _ureg.radian, "Exit face angle."),
        "HGAP": (0.0 * _ureg.m, "Magnet gap."),
        "FINTX": (0.0, "Exit fringe field integral."),
        "R2": (_np.inf * _ureg.m, "Entrance face curvature radius"),
    }

    @property
    def parameters(self) -> list:
        return list(
            map(
                float,
                [
                    self.length,  # 0
                    self.ANGLE.m_as("radian"),  # 1
                    self.K1.m_as("m**-2"),  # 2
                    self.E2.m_as("radian"),  # 3
                    self.HGAP.m_as("m"),  # 4
                    self.FINTX,  # 5
                    self.R2.m_as("m"),  # 6
                ],
            ),
        )


class DipEdge(Magnet):
    """
    Define a DipEdge element.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the DipEdge with their default values.

    Examples:
        >>> d1 = DipEdge('D1', H=1*_ureg.mm**-1)
        >>> d1 #doctest: +NORMALIZE_WHITESPACE
            DipEdge: {'NAME': 'D1',
                      'AT_ENTRY': <Quantity(0, 'meter')>,
                      'AT_CENTER': <Quantity(0, 'meter')>,
                      'AT_EXIT': <Quantity(0, 'meter')>,
                      'APERTYPE': None,
                      'APERTURE': [],
                      'KINEMATICS': None,
                      'H': <Quantity(1, '1 / millimeter')>,
                      'E1': <Quantity(0.0, 'radian')>,
                      'HGAP': <Quantity(0.0, 'meter')>,
                      'FINT': 0.0}
    """

    PARAMETERS = {
        "H": (0.0 * _ureg.m**-1, "Inverse of the curvature radius."),
        "E1": (0.0 * _ureg.radian, "Entrance face angle."),
        "HGAP": (0.0 * _ureg.m, "Magnet gap."),
        "FINT": (0.0, "Fringe field integral."),
    }

    @property
    def parameters(self) -> list:
        h = self.H.m_as("m**-1")
        e1 = self.E1.m_as("radian")
        hgap = self.HGAP.m_as("m")
        fint = self.FINT
        return list(Bend.compute_fringe(h, e1, hgap, fint))


class Solenoid(Magnet):
    pass


class Multipole(Magnet):
    """
    Define a Multipole magnet.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the Multipole with their default values.

    Examples:
        >>> m1 = Multipole('M1', L=10*_ureg.cm, K1=4*_ureg.m**-2, K2=0.1*_ureg.m**-3)
        >>> m1 #doctest: +NORMALIZE_WHITESPACE
            Multipole: {'NAME': 'M1',
                        'AT_ENTRY': <Quantity(0, 'meter')>,
                        'AT_CENTER': <Quantity(0, 'meter')>,
                        'AT_EXIT': <Quantity(0, 'meter')>,
                        'APERTYPE': None,
                        'APERTURE': [],
                        'KINEMATICS': None,
                        'L': <Quantity(10, 'centimeter')>,
                        'K1': <Quantity(4, '1 / meter ** 2')>,
                        'K2': <Quantity(0.1, '1 / meter ** 3')>}
    """

    PARAMETERS = {
        "L": (0.0 * _ureg.m, "Magnet length."),
        "K1": (0.0 * _ureg.m**-2, "Quadrupolar normalized gradient."),
        "K2": (0.0 * _ureg.m**-3, "Sextupolar normalized gradient."),
    }

    @property
    def parameters(self) -> list:
        return list(
            map(
                float,
                [
                    self.L.m_as("m"),
                    self.K1.m_as("m**-2"),
                    self.K2.m_as("m**-3"),
                ],
            ),
        )


class Sextupole(Magnet):
    """
    Define a Sextupole magnet.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the Sextupole with their default values.

    Examples:
        >>> s1 = Sextupole('S1', L=1*_ureg.m, K2=3*_ureg.m**-3)
        >>> s1 #doctest: +NORMALIZE_WHITESPACE
            Sextupole: {'NAME': 'S1',
            'AT_ENTRY': <Quantity(0, 'meter')>,
            'AT_CENTER': <Quantity(0, 'meter')>,
            'AT_EXIT': <Quantity(0, 'meter')>,
            'APERTYPE': None, 'APERTURE': [],
            'KINEMATICS': None,
            'L': <Quantity(1, 'meter')>,
            'K2': <Quantity(3, '1 / meter ** 3')>}
    """

    PARAMETERS = {
        "L": (0.0 * _ureg.m, "Magnet length."),
        "K2": (0.0 * _ureg.m**-3, "Sextupolar normalized gradient."),
    }

    @property
    def parameters(self) -> list:
        return list(
            map(
                float,
                [
                    self.L.m_as("m"),
                    self.K2.m_as("m**-3"),
                ],
            ),
        )


class Octupole(Magnet):
    pass


class Decapole(Magnet):
    pass


class Dodecapole(Magnet):
    pass


class Kicker(Magnet):
    """
    Define a Kicker element.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the Kicker with their default values.

    Examples:
        >>> k1 = Kicker('K1', L=10*_ureg.cm, HKICK=0.1, VKICK=-0.2)
        >>> k1 #doctest: +NORMALIZE_WHITESPACE
            Kicker: {'NAME': 'K1',
                     'AT_ENTRY': <Quantity(0, 'meter')>,
                     'AT_CENTER': <Quantity(0, 'meter')>,
                     'AT_EXIT': <Quantity(0, 'meter')>,
                     'APERTYPE': None, 'APERTURE': [],
                     'KINEMATICS': None,
                     'L': <Quantity(10, 'centimeter')>,
                     'HKICK': 0.1,
                     'VKICK': -0.2,
                     'TILT': <Quantity(0.0, 'radian')>}
    """

    PARAMETERS = {
        "L": (0.0 * _ureg.m, "Kicker length."),
        "HKICK": (0.0, "The momentum change in the horizontal plane."),
        "VKICK": (0.0, "The momentum change in the vertical plane."),
        "TILT": (
            0.0 * _ureg.radian,
            "The roll angle about the longitudinal axis. A positive angle represents a "
            "clockwise rotation of the kicker",
        ),
    }
    """Parameters of the element, with their default value and their descriptions."""

    @property
    def parameters(self) -> list:
        tilt = -self.TILT.m_as("radian")
        ct = _np.cos(tilt)
        st = _np.sin(tilt)
        hkick = self.HKICK
        vkick = self.VKICK
        _hkick = ct * hkick + st * vkick
        _vkick = ct * vkick - st * hkick
        return list(
            map(
                float,
                [
                    self.L.m_as("m"),
                    _hkick,
                    _vkick,
                ],
            ),
        )


class TKicker(Kicker):
    pass


class HKicker(Magnet):
    """
    Define a HKicker element.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the HKicker with their default values.

    Examples:
        >>> h1 = HKicker('H1', L=10*_ureg.cm, KICK=0.1)
        >>> h1 #doctest: +NORMALIZE_WHITESPACE
            HKicker: {'NAME': 'H1',
                     'AT_ENTRY': <Quantity(0, 'meter')>,
                     'AT_CENTER': <Quantity(0, 'meter')>,
                     'AT_EXIT': <Quantity(0, 'meter')>,
                     'APERTYPE': None, 'APERTURE': [],
                     'KINEMATICS': None,
                     'L': <Quantity(10, 'centimeter')>,
                     'KICK': 0.1,
                     'TILT': <Quantity(0.0, 'radian')>}
    """

    PARAMETERS = {
        "L": (0.0 * _ureg.m, "Kicker length."),
        "KICK": (0.0, "The momentum change."),
        "TILT": (
            0.0 * _ureg.radian,
            "The roll angle about the longitudinal axis. A positive angle represents a "
            "clockwise rotation of the kicker",
        ),
    }
    """Parameters of the element, with their default value and their descriptions."""

    @property
    def parameters(self) -> list:
        tilt = -self.TILT.m_as("radian")
        ct = _np.cos(tilt)
        st = _np.sin(tilt)
        kick = self.KICK
        hkick = ct * kick
        vkick = -st * kick
        return list(
            map(
                float,
                [
                    self.L.m_as("m"),
                    hkick,
                    vkick,
                ],
            ),
        )


class VKicker(Magnet):
    """
    Define a VKicker element.

    Attributes:
        PARAMETERS (dict): Dictionary containing the parameters of the VKicker with their default values.

    Examples:
        >>> v1 = VKicker('V1', L=10*_ureg.cm, KICK=-0.2)
        >>> v1 #doctest: +NORMALIZE_WHITESPACE
            VKicker: {'NAME': 'V1',
                     'AT_ENTRY': <Quantity(0, 'meter')>,
                     'AT_CENTER': <Quantity(0, 'meter')>,
                     'AT_EXIT': <Quantity(0, 'meter')>,
                     'APERTYPE': None, 'APERTURE': [],
                     'KINEMATICS': None,
                     'L': <Quantity(10, 'centimeter')>,
                     'KICK': -0.2,
                     'TILT': <Quantity(0.0, 'radian')>}
    """

    PARAMETERS = {
        "L": (0.0 * _ureg.m, "Kicker length."),
        "KICK": (0.0 * _ureg.radian, "The momentum change."),
        "TILT": (
            0.0 * _ureg.radian,
            "The roll angle about the longitudinal axis. A positive angle represents a "
            "clockwise rotation of the kicker",
        ),
    }
    """Parameters of the element, with their default value and their descriptions."""

    @property
    def parameters(self) -> list:
        tilt = -self.TILT.m_as("radian")
        ct = _np.cos(tilt)
        st = _np.sin(tilt)
        kick = self.KICK
        hkick = st * kick
        vkick = ct * kick
        return list(
            map(
                float,
                [
                    self.L.m_as("m"),
                    hkick,
                    vkick,
                ],
            ),
        )



from typing import Any, List, Tuple, Union

import numpy as np
import pandas as pd
from scipy.optimize import minimize


def compute_matrix(beam: pd.DataFrame = None) -> List[Union[np.ndarray, Any]]:
    """Return the transformation matrix"""

    def get_matrix(u, v):
        cov = np.cov(u, v)
        emit = np.sqrt(np.linalg.det(cov))
        beta = cov[0, 0] / emit
        alpha = -cov[0, 1] / emit

        return [[1, 0], [alpha, beta]]

    x = beam["x"].values
    xp = beam["xp"].values
    y = beam["y"].values
    yp = beam["yp"].values

    return [get_matrix(x, xp), get_matrix(y, yp)]


def normalize_phase_space(beam: pd.DataFrame = None, matrix_h: np.ndarray = None, matrix_v: np.ndarray = None):
    bnorm = beam.copy(deep=True)

    b = [beam["x"].values, beam["xp"].values]
    d = [beam["y"].values, beam["yp"].values]

    # Normalize the coordinates
    bnorm["x"], bnorm["xp"] = np.dot(matrix_h, b)
    bnorm["y"], bnorm["yp"] = np.dot(matrix_v, d)

    return bnorm


# 2D algorithm
def compute_quantile_2dim(data: pd.DataFrame = None) -> Tuple[float, pd.DataFrame]:
    """
    Compute the quantile in two dimensions: momentum
    Args:
        data: initial beam dataframe

    Returns:
        optimize_quant: optimised quantile
        beam: optimised beam

    """

    beam = data.copy()

    beam["ptrans"] = beam["xp"] ** 2 + beam["yp"] ** 2

    # Compute the quantile
    quantile = minimize(
        lambda xi: get_deviation_2dim(
            data=beam,
            quant=xi,
        ),
        0,
        method="Nelder-Mead",
        tol=1e-8,
    )

    optimize_quant = quantile.x[0]

    # Apply the quantile
    beam = beam[beam["ptrans"] <= beam["ptrans"].quantile(1 - optimize_quant)]

    return optimize_quant, beam


def compute_chisquare_2dim(
    data_x: np.array = None,
    data_y: np.array = None,
) -> float:
    """Computes the chi-square distribution"""

    u_c = (data_x - data_x.mean()) / data_x.std()
    v_c = (data_y - data_y.mean()) / data_y.std()
    p_c = u_c**2 + v_c**2

    return p_c.std()


def get_deviation_2dim(data: pd.DataFrame = None, quant: np.array = None) -> float:
    data_quantile = data[data["ptrans"] <= data["ptrans"].quantile(1 - quant[0])]
    data_x = data_quantile["xp"].values
    data_y = data_quantile["yp"].values
    deviation_chisquare = np.abs((compute_chisquare_2dim(data_x, data_y)) - 2)
    return deviation_chisquare


def get_cutted_data_2dim(data: pd.DataFrame = None, quantile: float = 0) -> pd.DataFrame:
    beam = data.copy()

    beam["ptrans"] = beam["xp"] ** 2 + beam["yp"] ** 2
    beam = beam[beam["trans"] <= beam["trans"].quantile(1 - quantile)]

    return beam


# 4D algorithm
def compute_quantile_4dim(beam: pd.DataFrame = None) -> Tuple[float, pd.DataFrame]:
    """
    Compute the quantile in 4 dimension
    Args:
        beam: initial beam dataframe

    Returns:
        optimize_quant: optimized quantile
        beam: optimized beam
    """

    # Compute the normalisation matrix
    matrix_h, matrix_v = compute_matrix(beam)

    # Normalize the coordinates
    beam = normalize_phase_space(beam, matrix_h, matrix_v)

    # Compute the transversal coordinates
    beam["trans"] = beam["x"] ** 2 + beam["y"] ** 2 + beam["xp"] ** 2 + beam["yp"] ** 2

    # Compute the quantile
    quantile = minimize(
        lambda xi: get_deviation_4dim(
            data=beam,
            quant=xi,
        ),
        0,
        method="Nelder-Mead",
        tol=1e-8,
        options={"disp": False, "fatol": 1e-12},
    )

    optimize_quant = quantile.x[0]

    # Apply the quantile
    beam = beam[beam["trans"] <= beam["trans"].quantile(1 - optimize_quant)]

    # Denormalize the coordinates, use the inverse of the matrix
    beam = normalize_phase_space(beam, np.linalg.inv(matrix_h), np.linalg.inv(matrix_v))

    return optimize_quant, beam


def compute_chisquare_4dim(
    data_x: np.array = None,
    data_xp: np.array = None,
    data_y: np.array = None,
    data_yp: np.array = None,
) -> float:
    """Computes the chi-square distribution"""

    xn = (data_x - data_x.mean()) / data_x.std()
    xpn = (data_xp - data_xp.mean()) / data_xp.std()

    yn = (data_y - data_y.mean()) / data_y.std()
    ypn = (data_yp - data_yp.mean()) / data_yp.std()

    p_c = xn**2 + xpn**2 + yn**2 + ypn**2
    return p_c.var()


def get_deviation_4dim(data: pd.DataFrame = None, quant=None) -> float:
    if quant is None:
        quant = [0]
    data_quantile = data[data["trans"] <= data["trans"].quantile(1 - quant[0])]

    # Compute the chi-square std
    data_x = data_quantile["x"].values
    data_xp = data_quantile["xp"].values
    data_y = data_quantile["y"].values
    data_yp = data_quantile["yp"].values
    deviation_chisquare = np.abs(
        (
            compute_chisquare_4dim(
                data_x,
                data_xp,
                data_y,
                data_yp,
            )
        )
        - 8.0,
    )
    return deviation_chisquare


def get_cutted_data_4dim(beam: pd.DataFrame = None, quantile: float = 0) -> pd.DataFrame:
    # Compute the normalisation matrix
    matrix_h, matrix_v = compute_matrix(beam)

    # Normalize the coordinates
    beam = normalize_phase_space(beam, matrix_h, matrix_v)

    # Compute the transversal coordinates and apply quantile
    beam["trans"] = beam["x"] ** 2 + beam["y"] ** 2 + beam["xp"] ** 2 + beam["yp"] ** 2
    beam = beam[beam["trans"] <= beam["trans"].quantile(1 - quantile)]

    # Denormalize the coordinates
    beam = normalize_phase_space(beam, np.linalg.inv(matrix_h), np.linalg.inv(matrix_v))

    return beam



"""
TODO
"""
import numpy as _np

from .. import ureg as _ureg


class ScatteringModelType(type):
    @staticmethod
    def t(pv: float, p1v1: float, **kwargs) -> float:
        return 0.0


class FermiRossi(metaclass=ScatteringModelType):
    """ """

    @staticmethod
    def t(pv: float, p1v1: float, **kwargs) -> float:
        """

        Args:
            pv:
            p1v1:
            **kwargs:

        Returns:

        """
        es = 15.0  # MeV
        chi_0 = kwargs["material"].radiation_length.m_as("cm")
        return (es / pv) ** 2 * (1 / chi_0)


class DifferentialHighland(metaclass=ScatteringModelType):
    """ """

    @staticmethod
    def length(x, chi0):
        """

        Args:
            x:
            chi0:

        Returns:

        """
        return x / chi0

    @staticmethod
    def f_dh(length):
        """

        Args:
            length:

        Returns:

        """
        return 0.970 * (1 + (_np.log(length) / 20.7)) * (1 + (_np.log(length) / 22.7))

    @staticmethod
    def t(pv: float, p1v1: float, **kwargs) -> float:
        """

        Args:
            pv:
            p1v1:
            **kwargs:

        Returns:

        """
        material = kwargs.get("material")
        es = 14.1  # MeV
        chi0 = material.radiation_length.m_as("cm")
        x = material.required_thickness(kinetic_energy_out=pv * _ureg.MeV, kinetic_energy_in=p1v1 * _ureg.MeV).m_as(
            "cm",
        )
        return DifferentialHighland.f_dh(DifferentialHighland.length(x, chi0)) * (es / pv) ** 2 * (1 / chi0)


class ICRU(metaclass=ScatteringModelType):
    """ """

    @staticmethod
    def t(pv: float, p1v1: float, **kwargs) -> float:
        """

        Args:
            pv:
            p1v1:
            **kwargs:

        Returns:

        """
        pass


class ICRUProtons(metaclass=ScatteringModelType):
    """ """

    @staticmethod
    def t(pv: float, p1v1: float, **kwargs) -> float:
        """

        Args:
            pv:
            p1v1:
            **kwargs:

        Returns:

        """
        material = kwargs["material"]
        es = 15.0  # MeV
        chi_s = material.scattering_length.m_as("cm")
        return (es / pv) ** 2 * (1 / chi_s)


class DifferentialMoliere(metaclass=ScatteringModelType):
    """ """

    @staticmethod
    def t(pv: float, p1v1: float, **kwargs) -> float:
        """

        Args:
            pv:
            p1v1:
            **kwargs:

        Returns:

        """
        material = kwargs["material"]
        es = 15.0  # MeV
        chi_s = material.scattering_length.m_as("cm")
        return DifferentialMoliere.f_dm(p1v1, pv) * (es / pv) ** 2 * (1 / chi_s)

    @staticmethod
    def f_dm(p1v1: float, pv: float):
        """

        Args:
            p1v1:
            pv:

        Returns:

        """
        if pv <= 0:
            raise ValueError("'pv' must be > 0.")
        if p1v1 <= 0:
            raise ValueError("'p1v1' must be > 0.")
        if p1v1 <= pv:
            raise ValueError("Initial 'p1v1' must be larger than final 'pv'.")
        return (
            0.5244
            + 0.1975 * _np.log10(1 - (pv / p1v1) ** 2)
            + 0.2320 * _np.log10(pv)
            - 0.0098 * _np.log10(pv) * _np.log10(1 - (pv / p1v1) ** 2)
        )



from typing import Tuple

from numba import njit, prange
from numba.typed import List as nList
from numpy import cos, cosh, sin, sinh, sqrt


@njit(fastmath=True)
def _apply_tilt_rotation(
    x: float,
    xp: float,
    y: float,
    yp: float,
    ct: float,
    st: float,
    sign: int,
) -> Tuple[float, float, float, float]:
    """

    Args:
        x:
        xp:
        y:
        yp:
        ct:
        st:
        sign:

    Returns:

    """
    x_ = ct * x + sign * st * y
    y_ = ct * y - sign * st * x
    px_ = ct * xp + sign * st * yp
    py_ = ct * yp - sign * st * xp
    return x_, px_, y_, py_


@njit(parallel=True, fastmath=True)
def track_madx_srotation(b1, b2, element_parameters: nList, global_parameters: nList):
    """

    Args:
        b1:
        b2:
        element_parameters:
        global_parameters:

    Returns:

    """
    tilt: float = element_parameters[0]
    if tilt < 1e-8:
        b2 = b1.copy()
        return b1, b2

    st = sin(tilt)
    ct = cos(tilt)
    for i in prange(b1.shape[0]):
        x = b1[i, 0]
        px = b1[i, 1]
        y = b1[i, 2]
        py = b1[i, 3]
        x_, px_, y_, py_ = _apply_tilt_rotation(x, px, y, py, ct, st, 1)
        b2[i, 0] = x_
        b2[i, 1] = px_
        b2[i, 2] = y_
        b2[i, 3] = py_
        b2[i, 4] = b1[i, 4]
        b2[i, 5] = b1[i, 5]

    return b1, b2


@njit(parallel=True, fastmath=True)
def track_madx_drift(b1, b2, element_parameters: nList, global_parameters: nList):
    """
    Track through a drift. This methods follows directly from the method implemented in MAD-X.

    .. note:: From MAD-X: `ttdrf` in `trrun.f90`

    Args:
        b1:
        b2:
        element_parameters:
        global_parameters:

    Returns:

    """
    length: float = element_parameters[0]
    beta: float = global_parameters[0]
    time_of_flight: bool = b1.shape[0] == 7
    for i in prange(b1.shape[0]):
        px = b1[i, 1]
        py = b1[i, 3]
        pt = b1[i, 5]

        lpz = length / sqrt(1.0 + 2.0 * pt / beta + pt**2.0 - px**2.0 - py**2.0)

        b2[i, 0] = b1[i, 0] + lpz * px
        b2[i, 1] = b1[i, 1]
        b2[i, 2] = b1[i, 2] + lpz * py
        b2[i, 3] = b1[i, 3]
        b2[i, 4] = b1[i, 4]
        b2[i, 5] = b1[i, 5]
        if time_of_flight:
            b2[i, 6] = b1[i, 6] + (length - (1.0 + beta * pt) * lpz) / beta

    return b1, b2


@njit(parallel=True, fastmath=True)
def track_madx_drift_paraxial(b1, b2, element_parameters: nList, global_parameters: nList):
    """
    Track through a drift using the paraxial approximation. Not used by MAD-X.

    Args:
        b1:
        b2:
        element_parameters:
        global_parameters:

    Returns:

    """
    length: float = element_parameters[0]
    beta: float = global_parameters[0]
    time_of_flight: bool = b1.shape[0] == 7
    for i in prange(b1.shape[0]):
        b2[i, 0] = b1[i, 0] + length * b1[i, 1]  # X
        b2[i, 1] = b1[i, 1]  # PX
        b2[i, 2] = b1[i, 2] + length * b1[i, 3]  # Y
        b2[i, 3] = b1[i, 3]  # PY
        b2[i, 4] = b1[i, 4]  # DPP
        b2[i, 5] = b1[i, 5]  # PT
        if time_of_flight:
            b2[i, 6] = b1[i, 6] + (length - (1.0 + beta * b1[i, 5]) * length) / beta

    return b1, b2


@njit(parallel=True, fastmath=True)
def track_madx_quadrupole(b1, b2, element_parameters: nList, global_parameters: nList):
    """
    Track through a (thick) quadrupole. This methods follows directly from the method implemented in MAD-X.

    The Hamiltonian is

    H = (1/2) K1 x^2 + (1/2) px^2/(delta + 1)

    .. note:: From MAD-X: `ttcfd` in `trrun.f90`

    Args:
        b1:
        b2:
        element_parameters:
        global_parameters:

    Returns:

    """
    length = element_parameters[0]
    k1 = element_parameters[1]
    tilt = element_parameters[2]
    st: float
    ct: float

    if k1 == 0:
        return track_madx_drift(b1, b2, element_parameters, global_parameters)

    if tilt != 0.0:
        st = sin(tilt)
        ct = cos(tilt)

    for i in prange(b1.shape[0]):
        delta_plus_1 = b1[i, 4] + 1
        x = b1[i, 0]
        xp = b1[i, 1] / delta_plus_1  # This is the key point to remember
        y = b1[i, 2]
        yp = b1[i, 3] / delta_plus_1  # This is the key point to remember

        if tilt != 0.0:
            x, xp, y, yp = _apply_tilt_rotation(x, xp, y, yp, ct, st, 1)

        k1_ = k1 / delta_plus_1  # This is the key point to remember
        if k1_ > 0:
            kl = sqrt(k1_) * length
            sx = sin(kl) / sqrt(k1_)
            cx = cos(kl)
            sy = sinh(kl) / sqrt(k1_)
            cy = cosh(kl)
        else:
            kl = sqrt(-k1_) * length
            sx = sinh(kl) / sqrt(-k1_)
            cx = cosh(kl)
            sy = sin(kl) / sqrt(-k1_)
            cy = cos(kl)

        x_ = cx * x + sx * xp
        xp_ = (-k1_ * sx * x + cx * xp) * delta_plus_1
        y_ = cy * y + sy * yp
        yp_ = (k1_ * sy * y + cy * yp) * delta_plus_1

        if tilt != 0.0:
            x_, xp_, y_, yp_ = _apply_tilt_rotation(x_, xp_, y_, yp_, ct, st, -1)

        b2[i, 0] = x_
        b2[i, 1] = xp_
        b2[i, 2] = y_
        b2[i, 3] = yp_
        b2[i, 4] = b1[i, 4]
        b2[i, 5] = b1[i, 5]

    return b1, b2


@njit(parallel=True, fastmath=True)
def track_madx_bend(b1, b2, element_parameters: nList, global_parameters: nList):
    """
    Track through a (thick) combined function bend. This methods follows directly from the method implemented in MAD-X.

    .. note:: From MAD-X: `ttcfd` in `trrun.f90`

    Args:
        b1:
        b2:
        element_parameters:
        global_parameters:

    Returns:

    """
    length: float = element_parameters[0]
    angle: float = element_parameters[1]
    k1: float = element_parameters[2]
    # k2 = element_parameters[3]
    tilt: float = element_parameters[4]
    h: float = element_parameters[5]
    k0: float = element_parameters[6]
    entrance_fringe_x: float = element_parameters[7]
    entrance_fringe_y: float = element_parameters[8]
    exit_fringe_x: float = element_parameters[9]
    exit_fringe_y: float = element_parameters[10]

    if angle == 0:
        quad_parameters = nList()
        quad_parameters.append(length)
        quad_parameters.append(k1)
        quad_parameters.append(0.0)
        quad_parameters.append(tilt)
        return track_madx_quadrupole(
            b1,
            b2,
            quad_parameters,
            global_parameters,
        )

    if tilt != 0.0:
        st: float = sin(tilt)
        ct: float = cos(tilt)

    for i in prange(b1.shape[0]):
        delta_plus_1: float = b1[i, 4] + 1.0
        x: float = b1[i, 0]
        xp: float = b1[i, 1]
        y: float = b1[i, 2]
        yp: float = b1[i, 3]

        # Apply magnet rotation
        if tilt != 0.0:
            xp, xp, y, yp = _apply_tilt_rotation(x, xp, y, yp, ct, st, 1)

        # Apply entrance fringe field
        xp += entrance_fringe_x * x
        yp += entrance_fringe_y * y

        # Body of the magnet
        k0_ = k0 / delta_plus_1
        k1_ = k1 / delta_plus_1
        # k2_ = k2 / delta_plus_1
        kx = k0_ * h + k1_
        ky = -k1_

        if kx > 0:
            klx = sqrt(kx) * length
            sx = sin(klx) / sqrt(kx)
            cx = cos(klx)
        elif kx < 0:
            klx = sqrt(-kx) * length
            sx = sinh(klx) / sqrt(-kx)
            cx = cosh(klx)
        else:
            sx = length
            cx = 1

        if ky > 0:
            kly = sqrt(ky) * length
            sy = sin(kly) / sqrt(ky)
            cy = cos(kly)
        elif ky < 0:
            kly = sqrt(-ky) * length
            sy = sinh(kly) / sqrt(-ky)
            cy = cosh(kly)
        else:
            sy = length
            cy = 1

        xp /= delta_plus_1
        yp /= delta_plus_1

        x_: float = cx * x + sx * xp
        xp_: float = ((-kx * x - k0_ + h) * sx + cx * xp) * delta_plus_1
        y_: float = cy * y + sy * yp
        yp_: float = (-ky * sy * y + cy * yp) * delta_plus_1

        if kx != 0.0:
            x_ = x_ + (k0_ - h) * (cx - 1.0) / kx
        else:
            x_ = x_ - (k0_ - h) * 0.5 * length**2

        # Apply exit fringe field
        xp_ += exit_fringe_x * x_
        yp_ += exit_fringe_y * y_

        # Apply magnet rotation
        if tilt != 0.0:
            x_, xp_, y_, yp_ = _apply_tilt_rotation(x_, xp_, y_, yp_, ct, st, -1)

        b2[i, 0] = x_
        b2[i, 1] = xp_
        b2[i, 2] = y_
        b2[i, 3] = yp_
        b2[i, 4] = b1[i, 4]
        b2[i, 5] = b1[i, 5]

    return b1, b2


@njit(parallel=True, fastmath=True)
def track_madx_dipedge(b1, b2, element_parameters: nList, global_parameters: nList):
    fringe_x: float = element_parameters[0]
    fringe_y: float = element_parameters[1]

    for i in prange(b1.shape[0]):
        x: float = b1[i, 0]
        xp: float = b1[i, 1]
        y: float = b1[i, 2]
        yp: float = b1[i, 3]

        # Apply entrance fringe field
        xp += fringe_x * x
        yp += fringe_y * y

        b2[i, 0] = x
        b2[i, 1] = xp
        b2[i, 2] = y
        b2[i, 3] = yp
        b2[i, 4] = b1[i, 4]
        b2[i, 5] = b1[i, 5]

    return b1, b2


@njit(parallel=True, fastmath=True)
def track_madx_kicker(b1, b2, element_parameters: nList, global_parameters: nList):
    element_parameters[0] /= 2
    hkick: float = element_parameters[1]
    vkick: float = element_parameters[2]

    # Half drift
    track_madx_drift(b1, b2, element_parameters[0:1], global_parameters)
    b1_copy = b1.copy()

    # Kick
    for i in prange(b1.shape[0]):
        b1[i, 0] = b2[i, 0]
        b1[i, 1] = b2[i, 1] + hkick
        b1[i, 2] = b2[i, 2]
        b1[i, 3] = b2[i, 3] + vkick
        b1[i, 4] = b2[i, 4]
        b1[i, 5] = b2[i, 5]

    # Half drift
    track_madx_drift(b1, b2, element_parameters[0:1], global_parameters)
    element_parameters[0] *= 2

    return b1_copy, b2



"""
TODO
"""
import os
import sys
from typing import List, Mapping, Optional

import numpy as _np
import pandas as _pd
import scipy.constants
import scipy.interpolate
import yaml
from georges_core import Kinematics as _Kinematics
from georges_core import ureg as _ureg
from georges_core.kinematics import ekin_to_pv as _ekin_to_pv
from scipy.integrate import quad as _quad

from .. import Q_ as _Q
from .mcs import DifferentialMoliere as _DifferentialMoliere
from .mcs import ScatteringModelType as _ScatteringModelType


class MaterialException(Exception):  # pragma: no cover
    """Exception raised for errors when using zgoubidoo.Sequence"""

    def __init__(self, m):
        self.message = m


def __bdsim_read_data(path: str) -> _pd.DataFrame:
    return _pd.read_csv(
        os.path.join(path, "bdsim", "data.csv"),
        delimiter=",",
        index_col="material",
    )


__STAR_COLUMNS: List[str] = ["K", "eS", "nS", "tS", "csda", "prange", "factor"]


def __star_read_data(path: str, material_name: str) -> _pd.DataFrame:
    return _pd.read_csv(
        os.path.join(path, "pstar", f"{material_name}.txt"),
        skiprows=7,
        delimiter=" ",
        names=__STAR_COLUMNS,
        index_col=False,
    )


def __srim_read_data(path: str, material_name: str, density: float) -> _pd.DataFrame:
    data = _pd.read_csv(
        os.path.join(path, "srim", f"{material_name}.txt"),
        skiprows=23,
        skipfooter=20,
        engine="python",
        delim_whitespace=True,
        names=[
            "energy",
            "energy_unit",
            "stopping_elec",
            "stopping_nuclear",
            "projected_range",
            "projected_range_unit",
            "longitudinal_straggling",
            "longitudinal_straggling_unit",
            "lateral_straggling",
            "lateral_straggling_unit",
        ],
    )
    units = {
        "A": 1e-10,
        "um": 1e-6,
        "mm": 1e-3,
        "m": 1,
        "keV": 1e-3,
        "MeV": 1,
        "GeV": 1000,
    }

    data["projected_range"] = data["projected_range"].astype(float)
    data["projected_range_meter"] = data.apply(
        lambda e: units[e["projected_range_unit"]] * density * 100 * e["projected_range"],
        axis=1,
    )
    data["energy"] = data["energy"].astype(float)
    data["energy_scale"] = data.apply(lambda e: units[e["energy_unit"]] * e["energy"], axis=1)
    data["stopping_elec"] = data["stopping_elec"].astype(float)
    data["stopping_nuclear"] = data["stopping_nuclear"].astype(float)

    data["stopping_elec"] *= 1000
    data["stopping_nuclear"] *= 1000
    return data


class RangeDefinitionType(type):
    pass


class CSDARange(metaclass=RangeDefinitionType):
    pass


class ProjectedRange(metaclass=RangeDefinitionType):
    pass


class ElementType(type):
    @property
    def atomic_a(cls) -> Optional[float]:
        if cls.material_data is None:  # pragma: no cover
            return None
        return cls.material_data["A"]

    @property
    def atomic_z(cls) -> Optional[int]:
        if cls.material_data is None:  # pragma: no cover
            return None
        return cls.material_data["Z"]

    @property
    def atomic_radiation_length(cls) -> _Q:
        """

        Args:
            material:

        Returns:

        """
        a: float = cls.atomic_a
        z: int = cls.atomic_z

        return 716.4 * a * (1 / (z * (z + 1) * (_np.log(287 / _np.sqrt(z))))) * _ureg("g / cm**2")

    @property
    def atomic_scattering_length(cls) -> _Q:
        """
        See "Techniques of Proton Radiotherapy:Transport Theory", B. Gottschalk, 2012.

        Args:
            material:

        Returns:

        """
        alpha: float = scipy.constants.alpha  # Fine structure constant
        avogadro: float = scipy.constants.Avogadro
        re: float = scipy.constants.physical_constants["classical electron radius"][0] * 100  # in cm!
        a: float = cls.atomic_a
        z: int = cls.atomic_z
        return (
            1
            / (alpha * avogadro * re**2 * z**2 * (2 * _np.log(33219 * (a * z) ** (-1 / 3)) - 1) / a)
            * _ureg("g / cm**2")
        )


class Element(metaclass=ElementType):
    def __init_subclass__(cls, material_data, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.material_data = material_data

    def __str__(self):
        return self.__class__.__name__.lower()

    def __eq__(self, y):
        return str(self) == str(y)


class CompoundType(type):
    @property
    def valid_data(cls):
        return (
            cls.material_data is not None
            and cls.projected_range is not None
            or (cls.projected_range is not None and cls.csda_range is not None)
        )

    @property
    def density(cls) -> _Q:
        if cls.material_data is None:  # pragma: no cover
            return None
        return cls.material_data["rho"] * _ureg("g /cm**3")

    @density.setter
    def density(cls, value) -> _Q:
        cls.material_data["rho"] = value * _ureg("g /cm**3")

    @property
    def radiation_length(cls) -> _Q:
        length = 0
        for element, fraction in cls.material_data["fractions"].items():
            element = getattr(sys.modules[__name__], "elements")[element.title()]
            length += fraction / element.atomic_radiation_length
        return 1 / (cls.density * length)

    @property
    def scattering_length(cls) -> _Q:
        length = 0
        for element, fraction in cls.material_data["fractions"].items():
            element = getattr(sys.modules[__name__], "elements")[element.title()]
            length += fraction / element.atomic_scattering_length
        return 1 / (cls.density * length)

    def range(
        cls,
        kinetic_energy: _ureg.Quantity,
        range_definition: RangeDefinitionType = ProjectedRange,
    ) -> Optional[_ureg.Quantity]:
        """

        Args:
            kinetic_energy:
            range_definition:

        Returns:

        """
        if not cls.valid_data:  # pragma: no cover
            return None
        energy = kinetic_energy.m_as("MeV")
        density = cls.density.m_as("g/cm**3")
        if range_definition == ProjectedRange:
            return _np.exp(cls.projected_range(_np.log(energy))) / density * _ureg.cm
        elif range_definition == CSDARange:
            return _np.exp(cls.csda_range(_np.log(energy))) / density * _ureg.cm
        else:
            raise MaterialException("'projected' or 'csda' arguments are mutually exclusive and one must be defined.")

    def solve_range(
        cls,
        range: _ureg.Quantity,
        range_definition: RangeDefinitionType = ProjectedRange,
    ) -> Optional[_Kinematics]:
        """

        Args:
            range:
            range_definition:

        Returns:

        """
        if not cls.valid_data:
            return None
        r = range.m_as("cm")
        density = cls.density.m_as("g/cm**3")
        if range_definition == ProjectedRange:
            return _Kinematics(
                _np.exp(cls.projected_range.solve(_np.log(r * density), extrapolate=False))[0] * _ureg.MeV,
                kinetic=True,
            )
        elif range_definition == CSDARange:
            return _Kinematics(
                _np.exp(cls.csda_range.solve(_np.log(r * density), extrapolate=False))[0] * _ureg.MeV,
                kinetic=True,
            )
        else:
            raise MaterialException("'projected' or 'csda' arguments are mutually exclusive and one must be defined.")

    def stopping(cls, thickness: _ureg.Quantity, kinetic_energy: _ureg.Quantity) -> _Kinematics:
        """

        Args:
            thickness:
            kinetic_energy:

        Returns:

        """
        if not cls.valid_data:
            return _Kinematics(kinetic_energy)
        return cls.solve_range(cls.residual_range(thickness, kinetic_energy))

    def residual_range(cls, thickness: _ureg.Quantity, kinetic_energy: _ureg.Quantity) -> Optional[_ureg.Quantity]:
        """

        Args:
            thickness:
            kinetic_energy:

        Returns:

        """
        if not cls.valid_data:
            return None
        return cls.range(kinetic_energy) - thickness

    def required_thickness(
        cls,
        kinetic_energy_out: _ureg.Quantity,
        kinetic_energy_in: _ureg.Quantity,
        range_definition: RangeDefinitionType = ProjectedRange,
    ):
        """

        Args:
            kinetic_energy_out:
            kinetic_energy_in:
            range_definition:

        Returns:

        """
        if not cls.valid_data:
            return None
        return cls.range(kinetic_energy_in, range_definition=range_definition) - cls.range(
            kinetic_energy_out,
            range_definition=range_definition,
        )

    def scattering(
        cls,
        kinetic_energy: _ureg.Quantity,
        thickness: _ureg.Quantity,
        model: _ScatteringModelType = _DifferentialMoliere,
        compute_a0: bool = True,
        compute_a1: bool = True,
        compute_a2: bool = True,
    ) -> Mapping[str, float]:
        """
        Compute the Fermi-Eyges parameters A0, A1, A2 and B (emittance).

        Args:
            kinetic_energy:
            thickness:
            model:
            compute_a0:
            compute_a1:
            compute_a2:

        Returns:

        """
        if not cls.valid_data:
            """TODO"""
            "Set a warning message"
            return {
                "A": (0.0, 0.0, 0.0),
                "B": 0.0,
                "TWISS_ALPHA": _np.nan,
                "TWISS_BETA": _np.nan,
                "TWISS_GAMMA": _np.nan,
            }
        thickness = thickness.m_as("cm")

        def integrand(
            u: float,
            initial_energy: _ureg.Quantity,
            thickness: float,
            material: ElementType,
            scattering_model: _ScatteringModelType,
            n: int,
        ):
            return (thickness - u) ** n * scattering_model.t(
                _ekin_to_pv(cls.stopping(u * _ureg.cm, initial_energy).ekin).m_as("MeV"),
                _ekin_to_pv(initial_energy).m_as("MeV"),
                material=material,
            )

        a = [0.0, 0.0, 0.0]
        if compute_a0:
            a[0] = _quad(integrand, 0, thickness, args=(kinetic_energy, thickness, cls, model, 0))[0]  # Order 0
        if compute_a1:
            a[1] = 1e-2 * _quad(integrand, 0, thickness, args=(kinetic_energy, thickness, cls, model, 1))[0]  # Order 1
        if compute_a2:
            a[2] = 1e-4 * _quad(integrand, 0, thickness, args=(kinetic_energy, thickness, cls, model, 2))[0]  # Order 2

        if compute_a0 and compute_a1 and compute_a2:
            b = _np.sqrt(a[0] * a[2] - a[1] ** 2)  # Emittance in m.rad
        else:
            b = 0.0

        return {
            "A": a,
            "B": b,
            "TWISS_ALPHA": -a[1] / b if b != 0.0 else 0.0,
            "TWISS_BETA": a[2] / b if b != 0.0 else 0.0,
            "TWISS_GAMMA": a[0] / b if b != 0.0 else 0.0,
        }

    def energy_dispersion(cls, energy: _ureg.Quantity) -> float:
        """
        TODO

        Args:
            energy:

        Returns:

        """
        if cls.bdsim_data is None:
            return 0.0

        c0 = cls.bdsim_data["energy_dispersion_c0"]
        c1 = cls.bdsim_data["energy_dispersion_c1"]
        c2 = cls.bdsim_data["energy_dispersion_c2"]
        c3 = cls.bdsim_data["energy_dispersion_c3"]
        c4 = cls.bdsim_data["energy_dispersion_c4"]
        c5 = cls.bdsim_data["energy_dispersion_c5"]
        energy = energy.m_as("MeV")
        return c5 * energy**5 + c4 * energy**4 + c3 * energy**3 + c2 * energy**2 + c1 * energy + c0

    def losses(cls, energy: _ureg.Quantity) -> float:
        """
        TODO

        Args:
            energy:

        Returns:

        """
        if cls.bdsim_data is None:
            return 1.0

        c0 = cls.bdsim_data["loss_c0"]
        c1 = cls.bdsim_data["loss_c1"]
        c2 = cls.bdsim_data["loss_c2"]
        c3 = cls.bdsim_data["loss_c3"]
        energy = energy.m_as("MeV")
        return c3 * energy**3 + c2 * energy**2 + c1 * energy + c0


class Compound(metaclass=CompoundType):
    def __init_subclass__(cls, csda_range_data, projected_range_data, material_data, bdsim_data, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.csda_range = csda_range_data
        cls.projected_range = projected_range_data
        cls.material_data = material_data
        cls.bdsim_data = bdsim_data

    def __str__(self):  # pragma: no cover
        return self.__class__.__name__.lower()

    def __eq__(self, y):  # pragma: no cover
        return str(self) == str(y)


def __initialize_materials_database():
    db_path = os.path.dirname(__file__)

    # Read all material definitions
    with open(os.path.join(db_path, "materials.yaml"), "r") as stream:
        materials_definitions = yaml.safe_load(stream)

    # Read P-Star data
    pstar_data_files = [
        os.path.splitext(f)[0]
        for f in os.listdir(os.path.join(db_path, "pstar"))
        if os.path.isfile(os.path.join(db_path, "pstar", f)) and os.path.splitext(f)[1] == ".txt"
    ]
    pstar = {m: __star_read_data(db_path, m) for m in pstar_data_files if materials_definitions["compounds"].get(m)}

    # Read SRIM data
    srim_data_files = [
        os.path.splitext(f)[0]
        for f in os.listdir(os.path.join(db_path, "srim"))
        if os.path.isfile(os.path.join(db_path, "srim", f)) and os.path.splitext(f)[1] == ".txt"
    ]
    srim = {
        m: __srim_read_data(
            db_path,
            m,
            materials_definitions["compounds"].get(m)["rho"],
        )
        for m in srim_data_files
        if materials_definitions["compounds"].get(m)
    }

    # Read BDSIM data
    bdsim_data = __bdsim_read_data(db_path)

    # Create the interpolated functions for the projected ranges
    projected_ranges_pstar = {
        k: scipy.interpolate.CubicSpline(_np.log(v["K"]), _np.log(v["prange"])) for k, v in pstar.items()
    }
    projected_ranges_srim = {
        k: scipy.interpolate.CubicSpline(_np.log(v["energy_scale"]), _np.log(v["projected_range_meter"]))
        for k, v in srim.items()
    }
    projected_ranges = {**projected_ranges_pstar, **projected_ranges_srim}

    # Create the interpolated functions for the CSDA ranges
    csda_ranges_pstar = {
        k: scipy.interpolate.CubicSpline(_np.log(v["K"]), _np.log(v["csda"])) for k, v in pstar.items()
    }
    csda_ranges = {**csda_ranges_pstar}

    # Dynamically create the material classes
    elements = {}
    for k, v in materials_definitions["elements"].items():
        elements[k.title()] = ElementType(
            k.title(),
            (Element,),
            {},
            material_data=v,
        )
        setattr(sys.modules[__name__], "elements", elements)
    for k, v in materials_definitions["compounds"].items():
        setattr(
            sys.modules[__name__],
            k.title(),
            CompoundType(
                k.title(),
                (Compound,),
                {},
                projected_range_data=projected_ranges.get(k, None),
                csda_range_data=csda_ranges.get(k, None),
                material_data=v,
                bdsim_data=bdsim_data.loc[k] if k in bdsim_data.index else None,
            ),
        )
